/**------------------------------------*------------------------------------**
    RELEASED TO THE PUBLIC DOMAIN BY CODEWORKS DEVELOPMENT, DECEMBER 1993.
    You may freely copy, distribute and reuse the code in this example. Codeworks disclaims any warranty of any kind, expressed or implied, as to its fitness for any particular use.
    Author:  Andrew Vyrros, av@codeworks.com
 **------------------------------------*------------------------------------**/
/**------------------------------------*------------------------------------**
    File:    Doodle.m
    Project: DynaDoodle
    Info:    Implementation of the Doodle class. Doodle is a subclass of View for creating arbitrary silly drawings. Doodle implements the basic behavior of setting foreground an background colors and loading a nib with custom controls. Subclasses, which are loaded from dynamic modules, provide the specific drawing functionality.
 **------------------------------------*------------------------------------**/

#import "Doodle.h"
#import <appkit/appkit.h>


@implementation Doodle


/**------------------------------------*------------------------------------**
    Method:  initFrame:
    Info:    Designated initializer, initializes new Doodle instance. Sets foregroundColor and backgroundColor to reasonable defaults.
    Return:  (id)self
 **------------------------------------*------------------------------------**/
- initFrame:(const NXRect *)f
{
  /* Connect to superclass. */
  [super initFrame:f];
  
  /* Set defaults. */
  foregroundColor = NX_COLORBLACK;
  backgroundColor = NX_COLORWHITE;
  customControlView = nil;

  return self;
}


/**------------------------------------*------------------------------------**
    Method:  free
    Info:    Frees all resources allocated by Doodle.
    Return:  (id)self
 **------------------------------------*------------------------------------**/
- free
{
  /*
   * Since customControlView is probably dissociated from the View hierarchy,
   * must free it explicitly. 
   */
  [customControlView free];

  return [super free];
}


/**------------------------------------*------------------------------------**
    Method:  customControlView
    Info:    Provides access to customControlView instance variable. This is a View that contains any special controls that a Doodle might need to adjust its appearance. Doodle expects this outlet to be connected to a View in a nib inside the module package. The nib should have the same name as the class--for instance the Star class uses Star.nib. This nib will be loaded when the customControlView is needed.
    Return:  (id)self
 **------------------------------------*------------------------------------**/
- customControlView
{
  /* If outlet has not been set yet, load nib. */
  if (!customControlView)
    [self loadNib];

  return customControlView;
}


/**------------------------------------*------------------------------------**
    Method:  loadNib
    Info:    Loads Doodle's nib with Doodle as owner. This lets the Doodle connect its customControlView outlet to a View in the nib. Subclasses should provide their own version of a nib in the module package. The nib should have the same name as the class--for instance the Star class uses Star.nib.
    Return:  (id)self, nil if there was a problem
 **------------------------------------*------------------------------------**/
- loadNib
{
  NXBundle           *myBundle;
  char                nibPath[MAXPATHLEN + 1];

  /*
   * Get bundle used to load this class. Note that we must use [self class]
   * rather than [Doodle class] to get the actual loaded subclass. 
   */
  myBundle = [NXBundle bundleForClass:[self class]];

  /*
   * Get path to Doodle's nib. NXBundle knows how to search .lproj
   * directories inside module packages. 
   */
  [myBundle getPath:nibPath forResource:[[self class] name] ofType:"nib"];

  /* Load nib. */
  if ([NXApp loadNibFile:nibPath owner:self withNames:NO fromZone:[self zone]])
  {
    /*
     * Notify self that nib was loaded. This gives subclasses a chance to set
     * instance variables based on settings of controls in nib. 
     */
    [self didLoadNib];
    return self;
  }
  else
    return nil;
}


/**------------------------------------*------------------------------------**
    Method:  didLoadNib
    Info:    Notifies Doodle that nib has been loaded. This gives subclasses a chance to perform any additional initialization that is dependent on the nib. (This cannot be done with awakeFromNib, because the Doodle is not in the nib, it is the nib's owner!) Typically, Doodle subclasses will use didLoadNib to set instance variables to the default settings of controls in the nib.
    Return:  (id)self
 **------------------------------------*------------------------------------**/
- didLoadNib
{
  return self;
}


/**------------------------------------*------------------------------------**
    Method:  drawSelf::
    Info:    Basic View drawing method. Just as a developement aid, Doodle fills its bounds with its backgroundColor. Subclasses should override this with their own desired drawing code.
    Return:  (id)self
 **------------------------------------*------------------------------------**/
- drawSelf:(const NXRect *)rects :(int)numRects
{
  NXSetColor(backgroundColor);
  NXRectFill(&bounds);

  return self;
}


/**------------------------------------*------------------------------------**
    Method:  setForegroundColor:, setBackgroundColor:
    Info:    Set values of foregroundColor/backgroundColor instance variables, then update Doodle.
    Return:  (id)self
 **------------------------------------*------------------------------------**/
- setForegroundColor:(NXColor)c
{
  foregroundColor = c;
  [self update];

  return self;
}


- setBackgroundColor:(NXColor)c
{
  backgroundColor = c;
  [self update];

  return self;
}


@end
