/**------------------------------------*------------------------------------**
    RELEASED TO THE PUBLIC DOMAIN BY CODEWORKS DEVELOPMENT, DECEMBER 1993.
    You may freely copy, distribute and reuse the code in this example. Codeworks disclaims any warranty of any kind, expressed or implied, as to its fitness for any particular use.
    Author:  Andrew Vyrros, av@codeworks.com
 **------------------------------------*------------------------------------**/
/**------------------------------------*------------------------------------**
    File:    DoodleBundle.m
    Project: DynaDoodle
    Info:    Implementation of the DoodleBundle class. DoodleBundle is a subclass of NXBundle designed specifically for managing dynamically loaded modules containing a single Doodle object. DoodleManager creates one DoodleBundle instance for each doodle module that it finds in the Library directories. When asked for its Doodle, a DoodleBundle will dynamically load its code file, create an instance of its Doodle subclass, and then return the instance to DoodleManager. DoodleBundle also provides localized strings describing its Doodle.
 **------------------------------------*------------------------------------**/

#import "DoodleBundle.h"
#import "Doodle.h"
#import <appkit/appkit.h>


@implementation DoodleBundle


/**------------------------------------*------------------------------------**
    Method:  initForDirectory:
    Info:    Designated initializer, initializes new DoodleBundle instance. Doesn't do much other than connect to superclass. Just here for clarity.
    Return:  (id)self
 **------------------------------------*------------------------------------**/
- initForDirectory:(const char *)path
{
  /* Connect to superclass. */
  [super initForDirectory:path];

  /* Clear out instance variables. Paranoia code. */
  doodleName = NULL;
  doodleDescription = NULL;
  doodle = nil;

  return self;
}


/**------------------------------------*------------------------------------**
    Method:  free
    Info:    Frees all resources allocated by DoodleBundle.
    Return:  (id)nil
 **------------------------------------*------------------------------------**/
- free
{
  doodle = [doodle free];
  return [super free];
}


/**------------------------------------*------------------------------------**
    Method:  doodleName
    Info:    Provides access to the doodleName instance variable. This string names the Doodle for the user. DoodleManager uses the string to create a PopUpList of available Doodle modules. DoodleBundle does not look up doodleName until needed.
    Note that we have DoodleBundle implement doodleName, rather than Doodle itself. This allows us to get the name of a module before we have actually loaded the module or instantiated the Doodle. By doing this, we can fill a UI control with the names of the options while still deferring loading and instantiating until the modules are needed.
    Instead of using the local string functions, we could have just taked the doodle module's file name and stripped the extension. However, that way user interface strings would be tied to file names. Also, our method lets us substitute more descriptive names--for instance, Face can become Smiley-Face.
    Return:  (const char *)doodleName
 **------------------------------------*------------------------------------**/
- (const char *)doodleName
{
  /* If doodleName is not set, look up. */
  if (!doodleName)
    /*
     * Note that there is some subtle trickery at work here. This call to
     * NXLocalizedStringFromTableInBundle() uses the key "Name" to look up
     * the value for the Doodle's name in the table Doodle.strings. This
     * relies on the fact each Doodle is in a separate DoodleBundle, with its
     * own Doodle.strings file. The file should have an entry under the key
     * "Name" with the proper full string. This isn't the usual way to use
     * NXLocalizedStringFromTableInBundle, but it works well here. 
     */
    doodleName = NXLocalizedStringFromTableInBundle("Doodle.strings", self, "Name", "Doodle", "Name of Doodle");

  return doodleName;
}


/**------------------------------------*------------------------------------**
    Method:  doodleDescription
    Info:    Provides access to the doodleDescription instance variable. This string describes the Doodle for the user. DoodleManager uses the string to fill a TextField. DoodleBundle does not look up doodleDescription until needed.
    The same points about doodleName apply to doodleDescription.
    Return:  (const char *)doodleDescription
 **------------------------------------*------------------------------------**/
- (const char *)doodleDescription
{
  if (!doodleDescription)
    doodleDescription = NXLocalizedStringFromTableInBundle("Doodle.strings", self, "Description", "Generic Doodle.", "Description of Doodle");

  return doodleDescription;
}


/**------------------------------------*------------------------------------**
    Method:  doodle
    Info:    Provides access to the doodle instance variable. Each DoodleBundle has its own Doodle instance that is created from a dynamically loaded Doodle subclass. DoodleManager lets DoodleBundle keep track of the actual Doodle. DoodleBundle will not load its code file or create its Doodle instance until it is asked. This way, DoodleManager can create the DoodleBundles when it scans the Library directories, but not load the code files or create Doodles until they are needed.
    Return:  (Doodle *)doodle
 **------------------------------------*------------------------------------**/
- (Doodle *)doodle
{
  Class               doodleClass;

  /* If doodle has not been created yet, */
  if (!doodle)
  {
    /* Load principal class, create instance. */
    doodleClass = [self principalClass];
    doodle = [[doodleClass allocFromZone:[self zone]] init];

    /*
     * Make sure that loaded doodle class is subclass of Doodle. Ideally, we
     * would like to do this before creating the instance, but Object does
     * not provide a method to check if a class is a subclass of another
     * class. (Of course, you could write your own method in a Category of
     * Object without too much trouble.) Instead, we check that the instance
     * is a kind of Doodle, and free it if not. 
     *
     * Using Protocols would be the other option here. That would free us to
     * created loaded classes that do not inherit from Doodle, as long as
     * they implement the required methods. In the scope of this example, it
     * is simpler to just use an abstract superclass. That also lets us have
     * Doodle implement common behavior to be inherited by the loaded
     * subclasses. 
     */
    if (![doodle isKindOf:[Doodle class]])
      doodle = [doodle free];
  }

  return doodle;
}


@end
