// MODocType.m
//
// by Mike Ferris
// Part of MOKit
// Copyright 1993, all rights reserved.

// ABOUT MOKit
// 
// MOKit is a collection of useful and general objects.  Permission is 
// granted by the author to use MOKit in your own programs in any way 
// you see fit.  All other rights to the kit are reserved by the author 
// including the right to sell these objects as part of a LIBRARY or as 
// SOURCE CODE.  In plain English, I wish to retain rights to these 
// objects as objects, but allow the use of the objects as pieces in a 
// fully functional program.  NO WARRANTY is expressed or implied.  The author 
// will under no circumstances be held responsible for ANY consequences to 
// you from the use of these objects.  Since you don't have to pay for 
// them, and full source is provided, I think this is perfectly fair.

#import "MOKit/MODocType.h"
#import "MOKit/MOString.h"
#import <objc/objc-runtime.h>

#define CLASS_VERSION						0
#define CLASS_NAME							"MODocType"

#define BUNDLE_TYPE							"bundle"

#define MOSTRING_CLASS_NAME					"MOString"

@interface MODocType(Private)

+ (Class)MO_loadClassBundle:(const char *)className;

@end

@implementation MODocType

static Class MOStringClass;

+ (Class)MO_loadClassBundle:(const char *)className
// Finds the bundle of the same name as the class, grabs it and loads the
// class from it and returns the named class.
{
	char pathBuff[MAXPATHLEN+1];
	id classBundle = nil;
	Class class = nil;
	
	// Load the bundle
	if ((class = objc_lookUpClass(className)) == nil)  {
		// class is not already loaded... load it.
		
		// Look for the bundle in the main bundle first, 
		// else try in this class's bundle.
		if (![[NXBundle mainBundle] getPath:pathBuff forResource:className 
					ofType:BUNDLE_TYPE])  {
			if (![[NXBundle bundleForClass:[self class]] getPath:pathBuff 
						forResource:className ofType:BUNDLE_TYPE])  {
				NXLogError("[%s loadClassBundle] failed to "
						"find %s class bundle.", [self name], className);
				return nil;
			}
		}
		classBundle = [[NXBundle allocFromZone:[self zone]] 
					initForDirectory:pathBuff];
		if (!classBundle)  {
			NXLogError("[%s loadClassBundle] failed to "
						"create bundle for class %s.", [self name], className);
			return nil;
		}
		if ((class = [classBundle classNamed:className]) == nil)  {
			NXLogError("[%s loadClassBundle] failed to "
						"load %s class from bundle.", [self name], className);
			return nil;
		}
	}
	
	return class;
}

+ initialize
// Set the version.
{
	if (self == objc_lookUpClass(CLASS_NAME))  {
		[self setVersion:CLASS_VERSION];

		// Load all the classes we need
		MOStringClass = [self MO_loadClassBundle:MOSTRING_CLASS_NAME];
	}
	return self;
}

- init
{
	return [self initForControllerClass:nil tag:-1 name:NULL extension:NULL 
				openSelector:NULL saveSelector:NULL];
}

- initForControllerClass:(Class)controller tag:(int)tag 
			name:(const char *)name extension:(const char *)ext 
			openSelector:(SEL)openSel saveSelector:(SEL)saveSel
// Designated initializer
{
	[super init];
	controllerClass = controller;
	typeTag = tag;
	typeName = [[MOStringClass allocFromZone:[self zone]] 
				initStringValue:name];
	typeExtension = [[MOStringClass allocFromZone:[self zone]] 
				initStringValue:ext];
	openSelector = openSel;
	saveSelector = saveSel;
	return self;
}

- free
{
	[typeName free];
	[typeExtension free];
	return [super free];
}

- (Class)controllerClass
{
	return controllerClass;
}

- (int)typeTag
{
	return typeTag;
}

- (const char *)typeName
{
	return [typeName stringValue];
}

- (const char *)typeExtension
{
	return [typeExtension stringValue];
}

- (BOOL)canOpenType
{
	return (openSelector!=NULL);
}

- (SEL)openSelector
{
	return openSelector;
}

- (BOOL)canSaveType
{
	return (saveSelector!=NULL);
}

- (SEL)saveSelector
{
	return saveSelector;
}


- read:(NXTypedStream *)strm
{
	int classVersion;

	[super read:strm];
	
	classVersion = NXTypedStreamClassVersion(strm, CLASS_NAME);
	
	switch (classVersion)  {
		case 0:		// First version.
			NXReadType(strm, "i", &typeTag);
			typeName = NXReadObject(strm);
			typeExtension = NXReadObject(strm);
			NXReadType(strm, ":", &openSelector);
			NXReadType(strm, ":", &saveSelector);
			break;
		default:
			NXLogError("[%s read:] class version %d cannot read "
						"instances archived with version %d", 
						CLASS_NAME, CLASS_VERSION, classVersion);
			typeTag = -1;
			typeName = [[MOStringClass allocFromZone:[self zone]] init];
			typeExtension = [[MOStringClass allocFromZone:[self zone]] init];
			openSelector = NULL;
			saveSelector = NULL;
			break;
	}
	return self;
}

- write:(NXTypedStream *)strm
{
	[super write:strm];
	NXWriteType(strm, "i", &typeTag);
	NXWriteObject(strm, typeName);
	NXWriteObject(strm, typeExtension);
	NXWriteType(strm, ":", &openSelector);
	NXWriteType(strm, ":", &saveSelector);
	return self;
}

