// MODocController.m
//
// by Mike Ferris
// Part of MOKit
// Copyright 1993, all rights reserved.

// ABOUT MOKit
// 
// MOKit is a collection of useful and general objects.  Permission is 
// granted by the author to use MOKit in your own programs in any way 
// you see fit.  All other rights to the kit are reserved by the author 
// including the right to sell these objects as part of a LIBRARY or as 
// SOURCE CODE.  In plain English, I wish to retain rights to these 
// objects as objects, but allow the use of the objects as pieces in a 
// fully functional program.  NO WARRANTY is expressed or implied.  The author 
// will under no circumstances be held responsible for ANY consequences to 
// you from the use of these objects.  Since you don't have to pay for 
// them, and full source is provided, I think this is perfectly fair.

#import "MOKit/MODocController.h"
#import "MOKit/MODocManager.h"
#import "MOKit/MOString.h"
#import "MOKit/MOPathString.h"
#import "MOKit/MOClassVariable.h"
#import <objc/objc-runtime.h>

#define CLASS_VERSION						0
#define CLASS_NAME							"MODocController"

#define BUNDLE_TYPE							"bundle"

#define MOSTRING_CLASS_NAME					"MOString"
#define MOPATHSTRING_CLASS_NAME				"MOPathString"
#define MODOCMANAGER_CLASS_NAME				"MODocManager"
#define MOCLASSVARIABLE_CLASS_NAME			"MOClassVariable"

#define UNTITLED_STRING						"Untitled"

@implementation MODocController

// True class variables via hashtables
static MOClassVariable *MO_classDocTypeName;
static MOClassVariable *MO_classOpenTypes;
static MOClassVariable *MO_classSaveTypes;

static Class MOStringClass;
static Class MOPathStringClass;
static Class MODocManagerClass;
static Class MOClassVariableClass;

+ (Class)MO_loadClassBundle:(const char *)className
// Finds the bundle of the same name as the class, grabs it and loads the
// class from it and returns the named class.
{
	char pathBuff[MAXPATHLEN+1];
	id classBundle = nil;
	Class class = nil;
	
	// Load the bundle
	if ((class = objc_lookUpClass(className)) == nil)  {
		// class is not already loaded... load it.
		
		// Look for the bundle in the main bundle first, 
		// else try in this class's bundle.
		if (![[NXBundle mainBundle] getPath:pathBuff forResource:className 
					ofType:BUNDLE_TYPE])  {
			if (![[NXBundle bundleForClass:[self class]] getPath:pathBuff 
						forResource:className ofType:BUNDLE_TYPE])  {
				NXLogError("[%s loadClassBundle] failed to "
						"find %s class bundle.", [self name], className);
				return nil;
			}
		}
		classBundle = [[NXBundle allocFromZone:[self zone]] 
					initForDirectory:pathBuff];
		if (!classBundle)  {
			NXLogError("[%s loadClassBundle] failed to "
						"create bundle for class %s.", [self name], className);
			return nil;
		}
		if ((class = [classBundle classNamed:className]) == nil)  {
			NXLogError("[%s loadClassBundle] failed to "
						"load %s class from bundle.", [self name], className);
			return nil;
		}
	}
	
	return class;
}

+ initialize
// Set the version.
{
	if (self == objc_lookUpClass(CLASS_NAME))  {
		[self setVersion:CLASS_VERSION];

		// Load all the classes we need
		MOStringClass = [self MO_loadClassBundle:MOSTRING_CLASS_NAME];
		MOPathStringClass = [self MO_loadClassBundle:MOPATHSTRING_CLASS_NAME];
		MODocManagerClass = [self MO_loadClassBundle:MODOCMANAGER_CLASS_NAME];
		MOClassVariableClass = [self MO_loadClassBundle:
					MOCLASSVARIABLE_CLASS_NAME];
		
		// Set up class variables
		MO_classDocTypeName = [[MOClassVariableClass allocFromZone:[self zone]] 
					initDoesFreeValues:YES];
		MO_classOpenTypes = [[MOClassVariableClass allocFromZone:[self zone]] 
					initDoesFreeValues:YES];
		MO_classSaveTypes = [[MOClassVariableClass allocFromZone:[self zone]] 
					initDoesFreeValues:YES];
	}
	return self;
}

+ setDocTypeName:(const char *)typeName
{
	id newVal;
	
	if (typeName)  {
		newVal = [[MOStringClass allocFromZone:[self zone]] 
				initStringValue:typeName];
	}  else  {
		newVal = nil;
	}
	[MO_classDocTypeName setObject:newVal forClass:self];
	return self;
}

+ (MOString *)docTypeName
{
	return [MO_classDocTypeName getObjectForClass:self];
}

+ addOpenType:(const char *)type
{
	if ((type) && (*type))  {
		List *typeList = [self openTypes];
	
		if (!typeList)  {
			typeList = [[List allocFromZone:[self zone]] init];
			[MO_classOpenTypes setObject:typeList forClass:self];
		}
	
		[typeList addObject:[[MOStringClass allocFromZone:[self zone]] 
					initStringValue:type]];
	}
	return self;
}

+ (List *)openTypes
{
	return (List *)[MO_classOpenTypes getObjectForClass:self];
}

+ addSaveType:(const char *)type
{
	if ((type) && (*type))  {
		List *typeList = [self saveTypes];
	
		if (!typeList)  {
			typeList = [[List allocFromZone:[self zone]] init];
			[MO_classSaveTypes setObject:typeList forClass:self];
		}
	
		[typeList addObject:[[MOStringClass allocFromZone:[self zone]] 
					initStringValue:type]];
	}
	return self;
}

+ (List *)saveTypes
{
	return (List *)[MO_classSaveTypes getObjectForClass:self];
}

+ startUnloading
// We have to deallocate the class variables.
{
	[MO_classDocTypeName free];
	[MO_classOpenTypes free];
	[MO_classSaveTypes free];
	
	return self;
}

- initFromFile:(const char *)path manager:aManager
{
	return [self initWithFrame:NULL fromFile:path manager:aManager];
}

- initWithFrame:(const char *)theFrameName
{
	return [self initWithFrame:theFrameName fromFile:NULL manager:nil];
}

- initWithFrame:(const char *)theFrameName 
			fromFile:(const char *)path manager:aManager
// Designated Initializer.  A null file means a new untitled doc.
{
	[super initWithFrameName:theFrameName];
	
	[self setManager:aManager];
	docName = [[MOPathStringClass allocFromZone:[self zone]] init];
	untitledString = [[MOPathStringClass allocFromZone:[self zone]] init];
	[self setFile:path];
	
	return self;
}

- free
// free our ivars
{
	[docName free];
	[untitledString free];
	if (manager)  [manager removeDocument:self];
	
	return [super free];
}

- nibDidLoad
// Set the window title, move the window according to the frameCycle.
{
	[super nibDidLoad];
	
	if (window)  {
		[self resetWindowTitle];
		if (manager)  {
			NXPoint pt;
			[manager getNextWindowLocation:&pt];
			[window moveTopLeftTo:pt.x :pt.y];
		}
	}
	return self;
}

- (BOOL)doSave
// override to do your actual saving into the file specified in docName.
{
	return YES;
}

- (BOOL)doRevert
// override this to actually read the file specified in docName
// into this document.
{
	return YES;
}

- (BOOL)doClear
// override this to totally clear this doc into a blank one.
{
	return YES;
}

- (BOOL)doPrint
// override this to print the document (including running print panel, etc...).
{
	return YES;
}

- saveAccessoryPopupAction:sender
{
	id savePanel = [SavePanel new];
	
	[savePanel setRequiredFileType:[[sender selectedCell] title]];
	
	return self;
}

- (BOOL)revert
// Called by DocManager when a file is opened or reverted.
{
	BOOL ret = YES;
	
	if ([docName isEmpty])  {
		ret = [self doClear];
	}  else  {
		ret = [self doRevert];
	}
	if (ret)  {
		[self resetWindowTitle];
		[self setDirty:NO];
	}
	return ret;
}

- (BOOL)save
// Called from various places when we are supposed to save.
{
	BOOL ret = YES;

	if ([docName isEmpty])  {
		return [self saveAs];
	}
	ret = [self doSave];
	if (ret)  {
		[self setDirty:NO];
	}
	return ret;
}

- (BOOL)saveAs
// Called by DocManager in response to Save As command.
{
	id savePanel = [SavePanel new];
	MOPathString *dir, *file;
	
	id saveTypes = [[self class] saveTypes];
	id saveAccessory = nil;
	const char *initialType = NULL;
	int i, c;
	
	if ([saveTypes count] == 0)  {
		return NO;
	}
	
	c = [saveTypes count];
	for (i=0; i<c; i++)  {
		const char *str = [[saveTypes objectAt:i] stringValue];
		if (str && *str)  {
			initialType = str;
			break;
		}
	}
	
	if (manager)  {
		saveAccessory = [manager getSaveAccessoryFor:self];
	}
	[savePanel setAccessoryView:saveAccessory];
	[savePanel setDelegate:self];
	[savePanel setTreatsFilePackagesAsDirectories:NO];
	if (!initialType)  {
		[savePanel setRequiredFileType:NULL];
	}  else  {
		[savePanel setRequiredFileType:initialType];
	}
	if ([docName isEmpty])  {
		dir = [[MOPathStringClass allocFromZone:[self zone]] 
					initPath:NXHomeDirectory()];
		file = [untitledString copy];
	}  else  {
		file = [docName file];
		dir = [docName directory];
	}
	if (![savePanel runModalForDirectory:[dir stringValue] 
				file:[file stringValue]])  {
		[file free];
		[dir free];
		return NO;
	}
	
	[docName setPath:[savePanel filename]];
	[self resetWindowTitle];
	[self save];
	
	[file free];
	[dir free];
	return YES;
}

- (BOOL)saveTo
// Called by DocManager in response to Save To command.  The bit about 
// the flushWindow is so the title bar doesn't flash the save to path and 
// then reset to the old path visibly.
{
	id copyString = [docName copy];
	BOOL copyDirty = [self isDirty];
	
	if (window)  {
		[window disableFlushWindow];
	}
	[self saveAs];
	
	[self setDirty:copyDirty];
	[docName free];
	docName = copyString;
	[self resetWindowTitle];
	
	if (window)  {
		[window reenableFlushWindow];
		[window flushWindow];
	}
	return YES;
}

- saveIfNeeded
// Save ourself only if we are dirty.
{
	if ([self isDirty])  [self save];
	return self;
}

- (BOOL)closeDoc
{
	[window performClose:self];
	return YES;
}

- (BOOL)print
{
	BOOL ret = [self doPrint];
	return ret;
}

- setManager:aManager
// Set the docManager that manages us.
{
	if (manager)  [manager removeDocument:self];
	manager = aManager;
	if (manager)  [manager addDocument:self];
	return self;
}

- manager
// Return our manager.
{
	return manager;
}

static BOOL isMOStringInList(MOString *string, List *list, int *pos)
{
	int i, c = [list count];
	for (i=0; i<c; i++)  {
		if ([string compare:[list objectAt:i]] == 0)  {
			if (pos)  *pos = i;
			return YES;
		}
	}
	return NO;
}

- setFile:(const char *)fName
// Name our file.  If fName is NULL, set up an untitled name.  Also, if
// the filename is not of a type we can save, make it untitled after loading.
{
	id extension;
	
	[docName setStringValue:fName];
	if ([docName isEmpty])  {
		if (manager)  {
			[untitledString setFromFormat:"%s-%d", UNTITLED_STRING, 
						[manager nextUntitledNum]];
		}  else  {
			[untitledString setFromFormat:"%s", UNTITLED_STRING];
		}
	}
	
	[self window:YES];
	if (window)  {
		[window disableFlushWindow];
	}
	if (![self revert])  {
		NXRunAlertPanel("Error", "Failed to revert document %s.", "OK", 
					NULL, NULL, [docName stringValue]);
	}
	
	// If we can't save files of whatever type this is, set us to untitled.
	// This must be done after the revert.
	extension = [docName fileExtension];
	if (!isMOStringInList(extension, [[self class] saveTypes], NULL))  {
		[docName setNull];
		if (manager)  {
			[untitledString setFromFormat:"%s-%d", UNTITLED_STRING, 
						[manager nextUntitledNum]];
		}  else  {
			[untitledString setFromFormat:"%s", UNTITLED_STRING];
		}
		[self resetWindowTitle];
	}
	[extension free];
	if (window)  {
		[window reenableFlushWindow];
		[window flushWindow];
	}
	[self showWindow:self];

	return self;
}

- (BOOL)isUntitled
// Returns YES if this doc has no file.
{
	return [docName isEmpty];
}

- (const char *)file
// Return our filename (or the untitled name if we have no file.
{
	if (![docName isEmpty])  {
		return [docName stringValue];
	}  else  {
		return [untitledString stringValue];
	}
}

- docName
{
	return docName;
}

- resetWindowTitle
// Make sure that our window title is correct.
{
	if (window)  [window setTitleAsFilename:[self file]];
	return self;
}

- setDirty:(BOOL)flag
// Mark ourselves as edited.
{
	isDirty = flag;
	if (window)  [window setDocEdited:flag];
	return self;
}

- (BOOL)isDirty
// Returns whether or not the doc is dirty.
{
	return isDirty;
}

- windowDidBecomeMain:sender
// Inform the manager that we should be the current document.
{
	if (manager)  [manager setCurrentDocument:self];
	return self;
}

- windowDidResignMain:sender
// Inform the manager that we should no longer be the current document.
{
	if (manager)  [manager setCurrentDocument:nil];
	return self;
}

- windowWillClose:sender
// If we are dirty, ask if we should save before closing and give the chance
// to cancel.
{
	int ret;
	
	if ([self isDirty])  {
		ret = NXRunAlertPanel("Close", "Document %s has unsaved changes.  "
					"Do you wish to save, close anyway, or cancel?", 
					"Save", "Close Anyway", "Cancel", [self file]);
		if (ret==NX_ALERTDEFAULT)  {
			// save
			if (![self save])  {
				return nil;
			}
		}  else if (ret==NX_ALERTALTERNATE)  {
			// close anyway
			// do nothing, just let it close
		}  else  {
			// cancel
			return nil;
		}
	}
	// if we get to here, we are closing
	[NXApp delayedFree:self];
	return self;
}

@end

