// MODocManager.m
//
// by Mike Ferris
// Part of MOKit
// Copyright 1993, all rights reserved.

// ABOUT MOKit
// 
// MOKit is a collection of useful and general objects.  Permission is 
// granted by the author to use MOKit in your own programs in any way 
// you see fit.  All other rights to the kit are reserved by the author 
// including the right to sell these objects as part of a LIBRARY or as 
// SOURCE CODE.  In plain English, I wish to retain rights to these 
// objects as objects, but allow the use of the objects as pieces in a 
// fully functional program.  NO WARRANTY is expressed or implied.  The author 
// will under no circumstances be held responsible for ANY consequences to 
// you from the use of these objects.  Since you don't have to pay for 
// them, and full source is provided, I think this is perfectly fair.

#import "MOKit/MODocManager.h"
#import "MOKit/MODocController.h"
#import "MOKit/MOString.h"
#import "MOKit/MOPathString.h"
#import <objc/objc-runtime.h>

#define CLASS_VERSION						0
#define CLASS_NAME							"MODocManager"

#define MOSTRING_CLASS_NAME					"MOString"
#define MOPATHSTRING_CLASS_NAME				"MOPathString"

#define BUNDLE_TYPE							"bundle"

#define QUITPANEL_NIBNAME					"MODocQuitPanel.nib"
#define QUITPANEL_REVIEW		0
#define QUITPANEL_SAVEALL		1
#define QUITPANEL_QUIT			2
#define QUITPANEL_CANCEL		3

#define SAVEACCESSORY_NIBNAME				"MODocSaveAccessory.nib"

#define MAX_FRAME_CYCLE				10
#define FRAME_ADJUST				25.0
#define FRAME_START_LOC_X			150.0
#define FRAME_START_LOC_Y			30.0  // top of screen to top of window

#define NEW_MENU_TITLE				"New"
#define OPEN_MENU_TITLE				"Open"


@interface MODocManager(Private)

- MO_resetOpenMenus;

@end


@implementation MODocManager

static Class MOStringClass;
static Class MOPathStringClass;

+ (Class)MO_loadClassBundle:(const char *)className
// Finds the bundle of the same name as the class, grabs it and loads the
// class from it and returns the named class.
{
	char pathBuff[MAXPATHLEN+1];
	id classBundle = nil;
	Class class = nil;
	
	// Load the bundle
	if ((class = objc_lookUpClass(className)) == nil)  {
		// class is not already loaded... load it.
		
		// Look for the bundle in the main bundle first, 
		// else try in this class's bundle.
		if (![[NXBundle mainBundle] getPath:pathBuff forResource:className 
					ofType:BUNDLE_TYPE])  {
			if (![[NXBundle bundleForClass:[self class]] getPath:pathBuff 
						forResource:className ofType:BUNDLE_TYPE])  {
				NXLogError("[%s loadClassBundle] failed to "
						"find %s class bundle.", [self name], className);
				return nil;
			}
		}
		classBundle = [[NXBundle allocFromZone:[self zone]] 
					initForDirectory:pathBuff];
		if (!classBundle)  {
			NXLogError("[%s loadClassBundle] failed to "
						"create bundle for class %s.", [self name], className);
			return nil;
		}
		if ((class = [classBundle classNamed:className]) == nil)  {
			NXLogError("[%s loadClassBundle] failed to "
						"load %s class from bundle.", [self name], className);
			return nil;
		}
	}
	
	return class;
}

+ initialize
// Set the version.
{
	if (self == [MODocManager class])  {
		[self setVersion:CLASS_VERSION];

		// Load all the classes we need
		MOStringClass = [self MO_loadClassBundle:MOSTRING_CLASS_NAME];
		MOPathStringClass = [self MO_loadClassBundle:MOPATHSTRING_CLASS_NAME];
	}
	return self;
}

- init
// Call the DI.
{
	return [self initDocumentClass:nil];
}

- initDocumentClass:(Class)docControllerClass;
// Init the manager with defaults.  If docControllerClass is non-nil it is
// added as the first of potential several document controller classes this
// manager will manage.
{
	const NXScreen *screen = NULL;
	
	[super init];
	
	docList = [[List allocFromZone:[self zone]] init];
	currentDoc = nil;
	docClassList = [[List allocFromZone:[self zone]] init];
	if (docControllerClass)  [docClassList addObject:docControllerClass];
	
	untitledCount = 0;
	frameCycle = 0;
	
	screen = [NXApp mainScreen];
	
	windowStartingLocation.x = FRAME_START_LOC_X;
	windowStartingLocation.y = screen->screenBounds.size.height - 
				FRAME_START_LOC_Y;
	
	return self;
}

- free
// free stuff we allocated.
{
	[[docList freeObjects] free];
	[docClassList empty];
	[self MO_resetOpenMenus];
	if (newSubmenuCell) [newSubmenuCell free];
	if (newMenu) [newMenu free];
	if (openSubmenuCell) [openSubmenuCell free];
	if (openMenu) [openMenu free];
	[docClassList free];
	
	return [super free];
}

- awakeFromNib
{
	if (docMenu)  {
		[NXApp setAutoupdate:YES];
		[docMenu setDelegate:self];
		[openCell setUpdateAction:@selector(menuUpdate:) forMenu:docMenu];
		[newCell setUpdateAction:@selector(menuUpdate:) forMenu:docMenu];
		[saveCell setUpdateAction:@selector(menuUpdate:) forMenu:docMenu];
		[saveAsCell setUpdateAction:@selector(menuUpdate:) forMenu:docMenu];
		[saveToCell setUpdateAction:@selector(menuUpdate:) forMenu:docMenu];
		[saveAllCell setUpdateAction:@selector(menuUpdate:) forMenu:docMenu];
		[revertCell setUpdateAction:@selector(menuUpdate:) forMenu:docMenu];
		[closeCell setUpdateAction:@selector(menuUpdate:) forMenu:docMenu];
	}
	[self MO_resetOpenMenus];
	return self;
}


- new:sender
// create a new untitled doc.
{
	id newDoc;
	if ([docClassList count] != 1)  {
		NXBeep();
		return nil;
	}
	newDoc = [[[docClassList objectAt:0] allocFromZone:[self zone]] 
				initFromFile:NULL manager:self];
	if (!newDoc)  {
		return nil;
	}
	
	return self;
}

- newFromDocumentClass:sender
{
	id newDoc = [[[docClassList objectAt:[sender selectedTag]] 
				allocFromZone:[self zone]] initFromFile:NULL manager:self];
	if (!newDoc)  {
		return nil;
	}
	return self;
}

static BOOL isMOStringInList(MOString *string, List *list, int *pos)
{
	int i, c = [list count];
	for (i=0; i<c; i++)  {
		if ([string compare:[list objectAt:i]] == 0)  {
			if (pos)  *pos = i;
			return YES;
		}
	}
	return NO;
}

- open:sender
// Allow the user to open a document from disk.  We must handle all 
// doc extensions supported by all doc controller classes.  If two 
// classes support reading the same extension, the first class which 
// supports that extension is used to open files with that extension.
{
	id openPanel = [OpenPanel new];

	int currentClass, docClassCount, currentType, typeCount;
	List *docOpenTypes = nil;

	// We'll fill this list with each open type supported by each doc 
	// controller class.  No duplicates, though.
	List *allTypes = [[List allocFromZone:[self zone]] init];
	// For open type [allTypes abjectAt:i], [classForTypes objectAt:i] will
	// be the doc controller class object to use to read the type.
	List *classForTypes = [[List allocFromZone:[self zone]] init];

	const char **types = NULL;
	const char *aType;
	
	// First, fill allTypes and classForTypes with info for all the extensions
	// we need to handle.
	docClassCount = [docClassList count];
	if (docClassCount == 0)  {
		[allTypes free];
		[classForTypes free];
		return nil;
	}
	for (currentClass = 0; currentClass < docClassCount; currentClass++)  {
		docOpenTypes = [[docClassList objectAt:currentClass] openTypes];
		typeCount = [docOpenTypes count];
		for (currentType = 0; currentType < typeCount; currentType++)  {
			MOString *typeStr = [docOpenTypes objectAt:currentType];
			if ((![typeStr isEmpty]) && (!isMOStringInList(typeStr, 
						allTypes, NULL)))  {
				// add it to the list and add the current class to 
				// the classForTypes list.
				[allTypes addObject:[docOpenTypes objectAt:currentType]];
				[classForTypes addObject:[docClassList objectAt:currentClass]];
			}
		}
	}
	
	// Now make a nasty array like the open panel wants.
	typeCount = [allTypes count];
	if (typeCount == 0)  {
		[allTypes free];
		[classForTypes free];
		return nil;
	}
	NX_MALLOC(types, char *, typeCount+1);
	for (currentType=0; currentType<typeCount; currentType++)  {
		aType = [[allTypes objectAt:currentType] stringValue];
		types[currentType] = aType;
	}
	types[typeCount] = NULL;
	
	// Now run the damn thing.
	[openPanel allowMultipleFiles:NO];
	[openPanel chooseDirectories:NO];
	[openPanel setAccessoryView:nil];
	[openPanel setDelegate:self];
	[openPanel setTreatsFilePackagesAsDirectories:NO];
	if ([openPanel runModalForTypes:types])  {
		// Now we need to open the damn thing with the right class.
		MOPathString *path = [[MOPathStringClass allocFromZone:[self zone]] 
					initPath:[openPanel filename]];
		MOPathString *extension = [path fileExtension];
		
		int index = -1;
		
		if (!isMOStringInList(extension, allTypes, &index))  {
			[allTypes free];
			[classForTypes free];
			NX_FREE(types);
			[path free];
			[extension free];
			return nil;
		}

		[self openDocument:[path stringValue] 
					withClass:[classForTypes objectAt:index]];
		
		[path free];
		[extension free];
	}
	
	[allTypes free];
	[classForTypes free];
	NX_FREE(types);
	return self;
}

- openFromDocumentClass:sender
{
	id openPanel = [OpenPanel new];
	id openTypes = [[docClassList objectAt:[sender selectedTag]] openTypes];
	int i, c, cursor;
	const char **types = NULL;
	const char *aType;
	
	c = [openTypes count];
	if (c == 0)  {
		return nil;
	}
	NX_MALLOC(types, char *, c+1);
	cursor = 0;
	for (i=0; i<c; i++)  {
		aType = [[openTypes objectAt:i] stringValue];
		if (aType && *aType)  {
			types[cursor++] = aType;
		}
	}
	types[cursor] = NULL;
	if (cursor == 0)  {
		NX_FREE(types);
		return nil;
	}
	
	[openPanel allowMultipleFiles:NO];
	[openPanel chooseDirectories:NO];
	[openPanel setAccessoryView:nil];
	[openPanel setDelegate:self];
	[openPanel setTreatsFilePackagesAsDirectories:NO];
	if ([openPanel runModalForTypes:types])  {
		[self openDocument:[openPanel filename] 
					withClass:[docClassList objectAt:[sender selectedTag]]];
	}
	
	NX_FREE(types);
	return self;
}

- openDocument:(const char *)path withClass:(Class)docClass;
// Open a new document and initialize it from the given file.
{
	id newDoc;
	
	newDoc = [[docClass allocFromZone:[self zone]] 
				initFromFile:path manager:self];
	
	return self;
}

- save:sender
// Tell the current doc to save.
{
	if (currentDoc)  {
		[currentDoc save];
	}
	return self;
}

- saveAs:sender
// Tell the current doc to save as.
{
	if (currentDoc)  {
		[currentDoc saveAs];
	}
	return self;
}

- saveTo:sender
// Tell the current doc to save to.
{
	if (currentDoc)  {
		[currentDoc saveTo];
	}
	return self;
}

- saveAll:sender
// Save all the open docs.
{
	int i, c = [docList count];
	id temp;
	
	for (i=0; i<c; i++)  {
		temp = [docList objectAt:i];
		[temp saveIfNeeded];
	}
	return self;
}

- revert:sender
// Tell the current doc to revert.
{
	if (currentDoc)  {
		if (NXRunAlertPanel("Revert", "Do you really want to revert to "
					"%s version?  This cannot be undone.", 
					"Revert", "Cancel", NULL, 
					([currentDoc isUntitled]?"blank":"previously saved")) == 
					NX_ALERTDEFAULT)  {
			[currentDoc revert];
		}
	}
	return self;
}

- close:sender
// Tell the current doc to close.
{
	if (currentDoc)  {
		[currentDoc closeDoc];
	}
	return self;
}

- print:sender
// Tell the current doc to close.
{
	if (currentDoc)  {
		[currentDoc print];
	}
	return self;
}

- getNextWindowLocation:(NXPoint *)pt
{
	pt->x = windowStartingLocation.x + (FRAME_ADJUST * frameCycle);
	pt->y = windowStartingLocation.y - (FRAME_ADJUST * frameCycle);
	frameCycle++;
	if (frameCycle>=MAX_FRAME_CYCLE)  frameCycle = 0;
	
	return self;
}

- (int)nextUntitledNum
{
	untitledCount++;
	return untitledCount;
}

- setWindowStartingLocation:(const NXPoint *)pt
{
	windowStartingLocation.x = pt->x;
	windowStartingLocation.y = pt->y;
	return self;
}

- getSaveAccessoryFor:sender
// This will prepare and return the saveAccessory view for the given sender 
// document.  If necessary it will load the accessory nib first.  It returns 
// nil if there is only a single save type since no accessory is needed then.
{
	id popup;
	int i, c;
	id saveTypes = [[sender class] saveTypes];
	id extension;
	
	c = [saveTypes count];
	if (c <= 1)  {
		return nil;
	}
	
	if (!saveAccessoryBox)  {
		if (![NXApp loadNibSection:SAVEACCESSORY_NIBNAME owner:self])  {
			NXLogError("%s: failed to load nib file %s.", 
					[[self class] name], SAVEACCESSORY_NIBNAME);
			return nil;
		}
		[saveAccessoryBox removeFromSuperview];
		[saveAccessoryPanel free];
	}
	
	popup = [saveAccessoryPopup target];
	
	// remove all but the first item in the list
	for (i=[popup count]-1; i>0; i--)  {
		[[popup removeItemAt:i] free];
	}
	
	// set the first item's title to ""
	[[[popup itemList] cellAt:0:0] setTitle:""];
	
	// add the new items
	for (i=0; i<c; i++)  {
		const char *str = [[saveTypes objectAt:i] stringValue];
		if (str && *str)  {
			[popup addItem:str];
		}
	}
	
	// remove the first item (which was still left from the old.
	[[popup removeItemAt:0] free];
	
	// set the target/action
	[popup setTarget:sender];
	[popup setAction:@selector(saveAccessoryPopupAction:)];
	
	// set the button title to the first item in the list
	if ([sender isUntitled])  {
	}  else  {
		extension = [[sender docName] fileExtension];
		if (![extension isEmpty])  {
			[saveAccessoryPopup setTitle:[extension stringValue]];
		}  else  {
			[saveAccessoryPopup setTitle:[[[popup itemList] 
						cellAt:0:0] title]];
		}
		[extension free];
	}
	
	return saveAccessoryBox;
}

- addDocument:(MODocController *)aDocument
// Add a new document to our list.
{
	[docList addObject:aDocument];
	
	return self;
}

- removeDocument:(MODocController *)aDocument
// Remove a document to our list.  (DocController calls this when the
// document is closing.)
{
	[docList removeObject:aDocument];
	if (currentDoc == aDocument)  {
		currentDoc = nil;
	}
	
	return self;
}

- (MODocController *)findDocumentForWindow:aWindow
// Find the controller that controls the given window (if any).
{
	int i, c = [docList count];
	id temp;
	
	for (i=0; i<c; i++)  {
		temp = [docList objectAt:i];
		if ([temp window] == aWindow)  {
			return temp;
		}
	}
	return nil;
}

- (List *)documentList
// Return the List object we use to store our docs.
{
	return docList;
}

- makeDocumentsPerform:(SEL)aMethod with:anArg
// Distribute a message to the documents.
{
	[docList makeObjectsPerform:aMethod with:anArg];
	
	return self;
}

- (BOOL)areDocumentsDirty
// Returns YES if at least one open document is dirty.
{
	int i, c = [docList count];
	
	for (i=0; i<c; i++)  {
		if ([[docList objectAt:i] isDirty])  {
			return YES;
		}
	}
	return NO;
}

- setCurrentDocument:(MODocController *)aDocument
// Sets the current doc.  Called from DocController's window delegate
// methods.
{
	currentDoc = aDocument;
	
	return self;
}

- currentDocument
// Returns the current doc's controller.
{
	
	return currentDoc;
}

- addDocumentClass:(Class)docControllerClass
// Add the class to the list of classes we manage.
{
	[docClassList addObject:docControllerClass];
	[self MO_resetOpenMenus];
	return self;
}

- removeDocumentClass:(Class)docControllerClass
{
	[docClassList removeObject:docControllerClass];
	[self MO_resetOpenMenus];
	return self;
}

- (List *)documentClassList
// Return the list of classes we manage.
{
	return docClassList;
}

- (int)docClassCount
{
	return [docClassList count];
}

- appWillTerminate:sender
// If there are unsaved docs, put up a panel asking if the user wants to
// Review Unsaved, Save All, Quit Anyway, or Cancel.
{
	int ret;
	int i, c;
	
	if ([self areDocumentsDirty])  {
		if (!quitPanel)  {
			if (![NXApp loadNibSection:QUITPANEL_NIBNAME owner:self])  {
				NXLogError("%s: failed to load nib file %s.", 
						[[self class] name], QUITPANEL_NIBNAME);
				return nil;
			}
		}
		ret = [NXApp runModalFor:quitPanel];
		[quitPanel orderOut:self];
		
		if (ret == QUITPANEL_REVIEW)  {
			c = [docList count];
			for (i=c-1;i>=0;i--)  {
				[[[docList objectAt:i] window] makeKeyAndOrderFront:self];
				if ([[docList objectAt:i] windowWillClose:[[docList 
							objectAt:i] window]] == nil)  {
					return nil;
				}
				[[[docList objectAt:i] window] orderOut:self];
			}
			return self;
		}  else if (ret == QUITPANEL_SAVEALL)  {
			[self saveAll:self];
			return self;
		}  else if (ret == QUITPANEL_QUIT)  {
			return self;
		}  else if (ret == QUITPANEL_CANCEL)  {
			return nil;
		}
	}
	return self;  // nil to abort termination
}

- quitPanelStopModalAction:sender
{
	[NXApp stopModal:[sender selectedTag]];
	return self;
}

#define QUITPANEL_OK_BUTTON_ROW		1
#define QUITPANEL_OK_BUTTON_COL		2

- windowDidBecomeKey:sender
// For the quit panel... make sure that the return icon is only there
// if the window is key.
{
	[[buttonMatrix cellAt:QUITPANEL_OK_BUTTON_ROW:QUITPANEL_OK_BUTTON_COL] 
				setIconPosition:NX_ICONRIGHT];
	[buttonMatrix setIcon:"NXreturnSign" 
				at:QUITPANEL_OK_BUTTON_ROW:QUITPANEL_OK_BUTTON_COL];
	return self;
}

- windowDidResignKey:sender
// For the quit panel... make sure that the return icon is only there
// if the window is key.
{
	[[buttonMatrix cellAt:QUITPANEL_OK_BUTTON_ROW:QUITPANEL_OK_BUTTON_COL] 
				setIconPosition:NX_TITLEONLY];
	[buttonMatrix setIcon:NULL 
				at:QUITPANEL_OK_BUTTON_ROW:QUITPANEL_OK_BUTTON_COL];
	return self;
}

- (BOOL)menuUpdate:menuCell
{
	BOOL enabled = [menuCell isEnabled];
	BOOL docsDirty;
	if ((menuCell == openCell) || (menuCell == newCell))  {
		if (([docClassList count] == 0) && (enabled))  {
			[menuCell setEnabled:NO];
			return YES;
		}  else if (!enabled)  {
			[menuCell setEnabled:YES];
			return YES;
		}
	}  else if ((menuCell == saveCell) || (menuCell == saveAsCell) || 
				(menuCell == saveToCell) || (menuCell == closeCell))  {
		if ((currentDoc == nil) && (enabled))  {
			[menuCell setEnabled:NO];
			return YES;
		}  else if ((currentDoc != nil) && (!enabled))  {
			[menuCell setEnabled:YES];
			return YES;
		}
	}  else if (menuCell == saveAllCell)  {
		docsDirty = [self areDocumentsDirty];
		if ((docsDirty) && (!enabled))  {
			[menuCell setEnabled:YES];
			return YES;
		}  else if ((!docsDirty) && (enabled))  {
			[menuCell setEnabled:NO];
			return YES;
		}
	}  else if (menuCell == revertCell)  {
		if ((currentDoc == nil) && (enabled))  {
			[menuCell setEnabled:NO];
			return YES;
		}  else if (currentDoc != nil)  {
			if (([currentDoc isDirty]) && (!enabled))  {
				[menuCell setEnabled:YES];
				return YES;
			}  else if ((![currentDoc isDirty]) && (enabled))  {
				[menuCell setEnabled:NO];
				return YES;
			}
		}
	}
	return NO;
}

- MO_resetOpenMenus
// Keeps the document menu in step with changes in the manager's class list.
{
	id itemList;
	int i, c;
	const char *name;
	id cell;
	int row, col;
	
	// See if the nib is loaded and we do menus
	if (!docMenu)  {
		return nil;
	}
	
	if ([docClassList count] > 1)  {
		// We should have submenus for new and open.
		
		// First do the newMenu
		if (!newMenu)  {
			newMenu = [[Menu allocFromZone:[Menu menuZone]] 
						initTitle:NEW_MENU_TITLE];
		}
		itemList = [newMenu itemList];
		[itemList renewRows:0 cols:1];
		c = [docClassList count];
		for (i=0; i<c; i++)  {
			name = [[[docClassList objectAt:i] docTypeName] stringValue];
			[itemList addRow];
			cell = [itemList cellAt:i:0];
			[cell setTitle:name];
			[cell setTag:i];
			[cell setTarget:self];
			[cell setAction:@selector(newFromDocumentClass:)];
		}
		[newMenu sizeToFit];
		
		// Now make a newSubmenuItem
		if (!newSubmenuCell)  {
			newSubmenuCell = [[MenuCell allocFromZone:[Menu menuZone]] 
						initTextCell:NEW_MENU_TITLE];
		}
		
		// Put the submenu item in the doc menu in place of the new item
		itemList = [docMenu itemList];
		if (![itemList getRow:&row andCol:&col ofCell:newSubmenuCell])  {
			[itemList getRow:&row andCol:&col ofCell:newCell];
			[itemList putCell:newSubmenuCell at:row :col];
			[docMenu setSubmenu:newMenu forItem:newSubmenuCell];
			[docMenu sizeToFit];
		}
		
		// Now do the open menu
		if (!openMenu)  {
			openMenu = [[Menu allocFromZone:[Menu menuZone]] 
						initTitle:OPEN_MENU_TITLE];
		}
		itemList = [openMenu itemList];
		[itemList renewRows:0 cols:1];
		c = [docClassList count];
		for (i=0; i<c; i++)  {
			name = [[[docClassList objectAt:i] docTypeName] stringValue];
			[itemList addRow];
			cell = [itemList cellAt:i:0];
			[cell setTitle:name];
			[cell setTag:i];
			[cell setTarget:self];
			[cell setAction:@selector(openFromDocumentClass:)];
		}
		[openMenu sizeToFit];
		
		// Now make an openSubmenuItem
		if (!openSubmenuCell)  {
			openSubmenuCell = [[MenuCell allocFromZone:[Menu menuZone]] 
						initTextCell:OPEN_MENU_TITLE];
		}
		
		// Put the submenu item in the doc menu below the open item
		itemList = [docMenu itemList];
		[itemList getRow:&row andCol:&col ofCell:openCell];
		if ([itemList cellAt:row+1:col] != openSubmenuCell)  {
			[itemList insertRowAt:row+1];
			[[itemList putCell:openSubmenuCell at:row+1 :col] free];
			[docMenu setSubmenu:openMenu forItem:openSubmenuCell];
			[docMenu sizeToFit];
		}
	}  else  {
		// We should not have submenus.
		
		itemList = [docMenu itemList];
		
		// take out the new submenu item if its there and put back the new item
		if ([itemList getRow:&row andCol:&col ofCell:newSubmenuCell])  {
			[itemList putCell:newCell at:row :col];
		}
		// take out the open submenu item if its there
		if ([itemList getRow:&row andCol:&col ofCell:openSubmenuCell])  {
			[itemList removeRowAt:row andFree:NO];
		}
		[docMenu sizeToFit];
		
	}
	return self;
}


@end
