//
//	MiscAppIconAnimator.m -- This is an Object to animate the icon
//				for an application.  This Object will allow
//				similiar activity to NeXTMail when new mail arrives.
//		Written by Craig Laurent Copyright 1995 by Craig Laurent.
//				Version 1.0.  All rights reserved.
//
//		This notice may not be removed from this source code.
//
//	This object is included in the MiscKit by permission from the author
//	and its use is governed by the MiscKit license, found in the file
//	"License.rtf" in the MiscKit distribution.  Please refer to that file
//	for a list of all applicable permissions and restrictions.
//	
//#import "MiscAppIconAnimator.h"
#import <misckit/MiscAppIconAnimator.h> 

@implementation MiscAppIconAnimator

/*  This is the callback function for the DPS Timed Entry. 
It in turn calls our animateIcon method.  DPS Timed Entries require functions not methods, as callbacks. */
void MiscAppIconAnimate (DPSTimedEntry timedEntry, double  now, void *who)
{
	[(id)who animateIcon];
}


- init
{
	return [self initWithImageArray:[[List alloc] init] animationPattern:[[List alloc] init] andPatternLoops:0];
}

//** designated initializer
- initWithImageArray:(List*)anArray animationPattern:(List*)patternArray andPatternLoops:(int)loops
{
	if (self = [super init]) {
		//** set default value for instance variables
		appTile = nil;
		appIconContentView = nil;
		imageArray = nil;
		animationPattern = nil;
		patternLoops = 0;
		timeInterval = 0.2;
		indexOfStopIcon = 0;
		indexOfEndIcon = 1;
		timedEntry = 0;
		patternIterator = 0;
		loopCount = 0;
		patternEnd = 0;
		offset = 0;

		//** set value for instance variables
		[self setImageArray:anArray];
		[self setAnimationPattern:patternArray];
		[self setPatternLoops:loops];
		
		return self;
	}
	return nil;
}


- free
{
    [imageArray free];
    [animationPattern free];
	return [super free];
}


//** instance methods
- (List*)imageArray
{ return imageArray; }
- (void)setImageArray:(List*)anArray
{
	[imageArray free];
	imageArray = anArray;
}

- (List*)animationPattern
{ return animationPattern; }
- (void)setAnimationPattern:(List*)anArray
{
	[animationPattern free];
	animationPattern = anArray;

	patternEnd = [animationPattern count];
}

- (int)patternLoops
{	return patternLoops; }
- (void)setPatternLoops:(int)numLoops
{	patternLoops = numLoops; }

- (float)timeInterval
{	return timeInterval; }
- (void)setTimeInterval:(float)interval
{	timeInterval = interval; }

- (int)indexOfEndIcon
{	return indexOfEndIcon; }
- (void)setIndexOfEndIcon:(int)endIcon
{	indexOfEndIcon = endIcon; }

- (int)indexOfStopIcon
{	return indexOfStopIcon; }
- (void)setIndexOfStopIcon:(int)stopIcon
{	indexOfEndIcon = stopIcon; }


//** methods to change Animation
/* startAnimation - starts running the animation sequence */
- (void)startAnimation:sender
{
//  if none exists, add a timed entry to begin the animation sequence...
	
	if (!timedEntry) {
	    timedEntry = DPSAddTimedEntry([self timeInterval], MiscAppIconAnimate, (void*)self, NX_MODALRESPTHRESHOLD);
	}
}

/* stopAnimation - stops the animation sequence and sets the application icon to a standard image */
- (void)stopAnimation
{
//  Indicate Animation stopped by setting image to image one.  Do necessary housekeeping.
	
	[self removeTimedEntry];
	if ([[self imageArray] count]) {
		[self displayImage: (NXImage*)[[self imageArray] objectAt: [self indexOfStopIcon]]];
	}
	patternIterator = 0;
	loopCount = 0;
}


//** Internal methods
/* removeTimedEntry - removes the timedEntry if it is running */
- (void)removeTimedEntry
{
	if (timedEntry) {
		DPSRemoveTimedEntry(timedEntry);
		timedEntry = 0;
	}
}

/* cornerPointForImage: -  calculates and returns the point that should be used as the corner for this image.  This is used to make sure the image is centered on the application icon. */
- (NXPoint)cornerPointForImage:(NXImage*)anImage
{
//*  You must center the composited image within
//*   the contentView of the appIcon window.
	NXSize	appTileSize;
	NXSize	imageSize;
	NXPoint	cornerPoint = { 0.0, 0.0 };

	[appTile getSize: &appTileSize];
	[anImage getSize: &imageSize];
	
	if ( imageSize.width < appTileSize.width ) 
		cornerPoint.x = (appTileSize.width - imageSize.width ) / 2.0;
	
	if ( imageSize.height < appTileSize.height ) 
		cornerPoint.y = (appTileSize.height - imageSize.height ) / 2.0;
		
	return cornerPoint;
}

/* displayImage: - displays the given image into the application icon. */
- (void)displayImage:(NXImage*)anImage
{
//*  NXAppTile is composited first at 0,0 of the icon window's content view 
//*  (this is required in order to maintain the NeXT icon look).
//*  'anImage' is then composited at center (centering is also a requirement).

	NXPoint	contentViewOrigin = { 0.0, 0.0 };
	NXPoint	cornerPoint;

	//** get application image every time
	appTile = [NXImage findImageNamed:"NXAppTile"];
	appIconContentView = [[NXApp appIcon] contentView];
	cornerPoint = [self cornerPointForImage:anImage];

	[appIconContentView lockFocus];
	[appTile composite:NX_SOVER toPoint:&contentViewOrigin];
	[anImage composite:NX_SOVER toPoint:&cornerPoint];
	[appIconContentView unlockFocus];
	[appIconContentView display];
}


/* displayImageAtIndex: - displays the image from the array, at the specified index into the application icon. */
- (void)displayImageAtIndex:(int)imageNum;
{
	if (imageNum < [[self imageArray] count])
		[self displayImage:[[self imageArray] objectAt:imageNum]];
}

/* animateIcon - called every time a timedEntry executes.  This contains logic for the icon animation (image sequencing).  The images contained in imageArray are used.  The animationPattern provides an index into that list, giving us the right image at the right time. */
- (void)animateIcon
{
	[self displayImage: (NXImage*)[[self imageArray] objectAt:(int)[[self animationPattern] objectAt:patternIterator++]]];

	if (patternIterator == (patternEnd - offset)) {
		[self removeTimedEntry];
		patternIterator = 0;
		loopCount++;
		if (loopCount < [self patternLoops]) {
			offset = 1;
//** the following only works for Object, not NSObject
			[self perform: @selector(startAnimation:) with:nil afterDelay:500 cancelPrevious:YES];
//			[self perform: @selector(startAnimation:) withObject:nil];
		} else {
			loopCount = offset = 0;
			[self displayImage:(NXImage*)[imageArray objectAt: [self indexOfEndIcon]]];
		}			
	}
}



//*********************
//** archiving methods
- write:(NXTypedStream *)stream
{
	int	count;
	int	entries;
	int temp;
	id	tempObj;

	[super write:stream];

	// write animationPattern
	entries = [animationPattern count];
	NXWriteType(stream, "i", &entries);
	for (count=0; count<entries; count++) {
		temp = (int)[animationPattern objectAt:count];
		NXWriteType(stream, "i", &temp);
	}

	// write imageArray
	entries = [imageArray count];
	NXWriteType(stream, "i", &entries);
	for (count=0; count<entries; count++) {
		tempObj = [imageArray objectAt:count];
		NXWriteType(stream, "@", &tempObj);
	}

	NXWriteType(stream, "i", &patternLoops);
	NXWriteType(stream, "f", &timeInterval);
	NXWriteType(stream, "i", &indexOfEndIcon);
	NXWriteType(stream, "i", &indexOfStopIcon);
	return self;
}

- read:(NXTypedStream *)stream
{
	int	count;
	int	entries;
	int temp;
	id	tempObj;

	[super read:stream];

	// read animationPattern
	animationPattern = [[List alloc] init];
	NXReadType(stream, "i", &entries);
	for (count=0; count<entries; count++) {
		NXReadType(stream, "i", &temp);
		[animationPattern addObject:(id)temp];
	}

	// read imageArray
	imageArray = [[List alloc] init];
	NXReadType(stream, "i", &entries);
	for (count=0; count<entries; count++) {
		NXReadType(stream, "@", &tempObj);
		[imageArray addObject:tempObj];
	}

	NXReadType(stream, "i", &patternLoops);
	NXReadType(stream, "f", &timeInterval);
	NXReadType(stream, "i", &indexOfEndIcon);
	NXReadType(stream, "i", &indexOfStopIcon);
	return self;
}


//****************************************
//** Distributed Object encoding methods
/* encodeRemotelyFor:.. -  Distributed Object method to indicate what should be passed, The object or a copy.  NSObject won't be passed as a copy. */
- encodeRemotelyFor:(NXConnection *)connection freeAfterEncoding:(BOOL *)flagp isBycopy:(BOOL)isBycopy
{
	if (isBycopy)
		return self;	//** return copy of object, not a proxy
//	return nil;	//** NSObject doesn't encode for DO.
	return [super encodeRemotelyFor:connection freeAfterEncoding:flagp isBycopy:isBycopy];
}

/* encodeUsing: -  Distributed Object method to encode the data for transfer */
- encodeUsing:(id <NXEncoding>)portal
{
	[portal encodeData:&animationPattern ofType:"@"];
	[portal encodeData:&imageArray ofType:"@"];
	[portal encodeData:&patternLoops ofType:"i"];
	[portal encodeData:&timeInterval ofType:"f"];
	[portal encodeData:&indexOfEndIcon ofType:"i"];
	[portal encodeData:&indexOfStopIcon ofType:"i"];
	return self;
}

/* decodeUsing: -  Distributed Object method to decode the data after transfer */
- decodeUsing:(id <NXDecoding>)portal
{
	self = [self init];

	//** decode data
	[portal decodeData:&animationPattern ofType:"@"];
	patternEnd = [animationPattern count];	// done in set method
	[portal decodeData:&imageArray ofType:"@"];
	[portal decodeData:&patternLoops ofType:"i"];
	[portal decodeData:&timeInterval ofType:"f"];
	[portal decodeData:&indexOfEndIcon ofType:"i"];
	[portal decodeData:&indexOfStopIcon ofType:"i"];

	return self;
}


@end
