/*============================= MiscBackWindow.m ===========================*/
//		Written by Dale Amon, copyright 1994 by Genesis Project, Ltd.
//				Version 0.8.  All rights reserved.
//
//		This notice may not be removed from this source code.
//
//	This object is included in the MiscKit by permission from the author
//	and its use is governed by the MiscKit license, found in the file
//	"LICENSE.rtf" in the MiscKit distribution.  Please refer to that file
//	for a list of all applicable permissions and restrictions.
//	

/* Class that impliments a full screen window with no title bars or other
   ornamentation visible. It can be set to a user selectable tier and will
   notify a delegate when it changes it's tier. It's content view is a
   MiscFramingView.

HISTORY
12-Jul-93  Dale Amon at GPL
	   Added basic archive methods

7-Feb-93   Dale Amon at GPL
	   Created. Based loosely on Backspace demo code.

NOTES:		There is a kludge to handle what appears to be a problem
		with the setting of the Window delegate.

		The awake method may be incomplete. I have not tested
		archiving at all.

		Assumes all screens are the same size as the main screen,
		which I don't like, so we'll just have to see if it still
		works if and when we get a bunch of screens of different sizes.
		ie, what happens when we go into the PC world?

*/

#import <misckit/misckit.h>
#import "MiscBWPS.h"

#define BACKGROUND_COLOR	NX_BLACK

@implementation MiscBackWindow

/*===========================================================================*/
/* Define localized strings */
/*===========================================================================*/

#define ILLEGAL_METHOD NXLocalizedStringFromTable("MiscKit", "  %s objects should not be sent '%s' messages\n", NULL, "Programmer error, not allowed to use the named method")

/*===========================================================================*/
/* Class methods */
/*===========================================================================*/
/* Initialize the class */

+ initialize {[MiscBackWindow setVersion:MISC_BACKWIN_VERSION_ID]; return self;}


/*===========================================================================*/

+ getFrameRect:(NXRect *)fRect forContentRect:(const NXRect *)cRect
	 style:(int)aStyle
 {
	fRect->origin.x=fRect->origin.y=0;
	[NXApp getScreenSize:&(fRect->size)];
	return self;
 }

/*---------------------------------------------------------------------------*/

+ getContentRect:(NXRect *)cRect forFrameRect:(const NXRect *)fRect
	   style:(int)aStyle
 {
	cRect->origin.x=cRect->origin.y=0;
	[NXApp getScreenSize:&(cRect->size)];
	return self;
 }

/*---------------------------------------------------------------------------*/

+ (NXCoord)minFrameWidth:(const char *)aTitle forStyle:(int)aStyle
	      buttonMask:(int)aMask
 {
	NXSize s;
	[NXApp getScreenSize:&s];
	return s.width;
 }


/*===========================================================================*/
/* Init, alloc and free methods */
/*===========================================================================*/
/* Block the designated initializer of our parent class */

- initContent:(const NXRect *)contentRect
	style:      (int) aStyle
	backing:    (int) backingType
	buttonMask: (int) mask
	defer:      (BOOL)flag
 {	[self error: ILLEGAL_METHOD, [[self class] name], sel_getName(_cmd)];
	return self;
 }


/*---------------------------------------------------------------------------*/
/* Initialize a window that is the size of the screen, has no borders and is
   in the background tier.
*/

- init
  {	NXRect	screenRect;
	id	content;

	screenRect.origin.x = screenRect.origin.y = 0;
	[NXApp getScreenSize:&(screenRect.size)];

	/* I selected buffered simply because that was what I saw
	   in BackSpace Controller.m */
	[super initContent:&screenRect
	       style:NX_TOKENSTYLE
	       backing:NX_BUFFERED
	       buttonMask:0
	       defer:NO];

	[self setBackgroundGray:BACKGROUND_COLOR];

	/* Replace current content with a MiscFramingView and free the old one */
	content    = [[MiscFramingView allocFromZone: [self zone]] init];
	[content setOutline:MISC_FV_PLAIN];
	[content setBackgroundGray:	BACKGROUND_COLOR];
	[[self setContentView: content] free];

	currentTier = MISC_BACKGROUNDTIER;
	_MiscTweakWindow([self windowNum], MISC_BACKGROUNDTIER);

	[self display];
	return self;
  }

/*===========================================================================*/
/* KLUDGE ALERT!
   This is a hack until I figure out what is going on. After either
   [self setDelegate: anObject]] or a [super setDelegate: anObject]
   delegate is still nil, ie 0. So I have overridden the Window method
   here to force the issue. I haven't a CLUE why it doesn't work the way
   it is supposed to...
*/

- setDelegate: anObject {delegate = anObject; return self;}


/*===========================================================================*/
/* Set the window tier, if it has changed, and notify the Window delegate */

- setWindowTier: (int) tier
 {
	if (tier != currentTier) {
		currentTier = tier;
		_MiscSetWindowLevel([self windowNum], tier);
#ifdef DEBUG
		NXLogError("Changing tier to %d", tier);
#endif
		if ([delegate respondsTo:@selector(windowDidChangeTierTo:)])
			[delegate windowDidChangeTierTo: tier];
		[self windowDidChangeTierTo: tier];
	}
	return self;
  }

- windowDidChangeTierTo:(int)aTier { return self; }

/*---------------------------------------------------------------------------*/
/* Put the full screen image in front of everything. */

- makeFrontdrop:sender
 {	[self setWindowTier: MISC_SAVERTIER];
	[self addToEventMask: NX_MOUSEMOVEDMASK];
	return self;
  }


/*---------------------------------------------------------------------------*/
/* Put the full screen image in back of everything. */

- makeBackdrop:sender
 {	[self setWindowTier: MISC_BACKGROUNDTIER];
	[self removeFromEventMask: NX_MOUSEMOVEDMASK];
	return self;
 }


// Debug scaffolding
//- (BOOL)acceptsFirstResponder	   {NXLogError("Window Set first responder");
//				    return YES;}
//
//- keyDown:    (NXEvent *) eventPtr {NXLogError("Window Key down");
//				    return self;}
//- mouseDown:  (NXEvent *) eventPtr {NXLogError("Window Mouse Down");
//				    return self;}
//- mouseMoved: (NXEvent *) eventPtr {NXLogError("Window Mouse Moved");
//				    return self;}

/*===========================================================================*/
/* Archive methods */
/*===========================================================================*/

- write: (NXTypedStream *) stream
  {	[super write: stream];
	NXWriteType (stream, "i", &currentTier);
	return self;
  }


- read: (NXTypedStream *) stream
  {	[super read: stream];
	NXReadType  (stream, "i", &currentTier);
	return self;
  }

/*---------------------------------------------------------------------------*/
/* play games with the tier number so that we can correctly reset the window
   status in the window manager. It might seems strange, but it gets by
   the check that ignores window tier values unless they have changed from
   the previous setting.

   NOTE: Will I have to handle the setting of the event mask? Will the
    	 delegate problem still occur?
*/

- awake
  {	int	tmp;

	[super awake];
	tmp = currentTier++;
	[self setWindowTier: tmp];
	return self;
  }

@end
