//
//	MiscDocument.m -- document controller
//		Written by Mike Ferris Copyright (c) 1993, 1995 by Mike Ferris.
//		Modified for use in the MiscKit by Don Yacktman from MODocController.
//				Version 1.0.  All rights reserved.
//		This notice may not be removed from this source code.
//
//	This object is included in the MiscKit by permission from the author
//	and its use is governed by the MiscKit license, found in the file
//	"LICENSE.rtf" in the MiscKit distribution.  Please refer to that file
//	for a list of all applicable permissions and restrictions.
//	

#import <misckit/MiscDocument.h>
#import <misckit/MiscDocType.h>
#import <misckit/MiscDocManager.h>
#import <misckit/MiscString.h>
#import <misckit/MiscClassVariable.h>
#import <objc/objc-runtime.h>

#define CLASS_VERSION					0
#define CLASS_NAME						"MiscDocument"

#define SAVEACCESSORY_NIBNAME			"MiscDocSaveAccessory.nib"

#define UNTITLED_STRING				NXLocalizedStringFromTableInBundle("MiscDocMenu", MiscKitBundle, "UntitledString", "Untitled", Title of a untitled window used by MiscDocument)

#define CLOSEALERTPANEL_TITLE			NXLocalizedStringFromTableInBundle("MiscDocAlertPanels", MiscKitBundle, "CloseTitle", "Close", Title of the Close alert panel used by MiscDocument)

#define CLOSEALERTPANEL_TEXT			NXLocalizedStringFromTableInBundle("MiscDocAlertPanels", MiscKitBundle, "CloseText", "Document %s has unsaved changes.  Do you wish to save, close anyway, or cancel?", Text of the Close alert panel used by MiscDocument)

#define CLOSEALERTPANEL_SAVE			NXLocalizedStringFromTableInBundle("MiscDocAlertPanels", MiscKitBundle, "CloseSaveButton", "Save", Title of the NX_ALERTDEFAULT button in the Close alert panel used by MiscDocument)

#define CLOSEALERTPANEL_CLOSEANYWAY		NXLocalizedStringFromTableInBundle("MiscDocAlertPanels", MiscKitBundle, "CloseCloseButton", "Close Anyway", Title of the NX_ALERTALTERNATE button in the Close alert panel used by MiscDocument)

#define CLOSEALERTPANEL_CANCEL			NXLocalizedStringFromTableInBundle("MiscDocAlertPanels", MiscKitBundle, "CloseCancelButton", "Cancel", Title of the NX_ALERTALTERNATE button in the Close alert panel used by MiscDocument)

#define ERRORALERTPANEL_TITLE			NXLocalizedStringFromTableInBundle("MiscDocAlertPanels", MiscKitBundle, "ErrorTitle", "Error", Title of the Error alert panel used by MiscDocument)

#define ERRORALERTPANEL_TEXT			NXLocalizedStringFromTableInBundle("MiscDocAlertPanels", MiscKitBundle, "ErrorText", "Failed to revert document %s.", Text of the Error alert panel used by MiscDocument)

#define ERRORALERTPANEL_OK			NXLocalizedStringFromTableInBundle("MiscDocAlertPanels", MiscKitBundle, "ErrorOKButton", "OK", Title of the NX_ALERTDEFAULT button in the Error alert panel used by MiscDocument)

@implementation MiscDocument

// True class variables via hashtables
static MiscClassVariable *Misc_classControllerName;
static MiscClassVariable *Misc_classDocTypes;

static id MiscKitBundle = nil;

+ initialize
// Set the version.  Load classes.  Load the MiscKit bundle.  
// Init class variables.
{
	if (self == objc_lookUpClass(CLASS_NAME))  {
		[self setVersion:CLASS_VERSION];		
		
		// Set up class variables
		Misc_classControllerName = [[MiscClassVariable 
					allocFromZone:[self zone]] 
					initDoesFreeValues:YES];
		Misc_classDocTypes = [[MiscClassVariable 
					allocFromZone:[self zone]] 
					initDoesFreeValues:YES];
	}
	return self;
}

+ setControllerName:(const char *)typeName
// Sets the name of this document class type.  This is used in New 
// and Open submenus that MiscDocManager creates if there is more 
// than one document type being managed.
{
	id newVal;
	
	if (typeName)  {
		newVal = [[MiscString allocFromZone:[self zone]] 
				initString:typeName];
	}  else  {
		newVal = nil;
	}
	[Misc_classControllerName setObject:newVal forClass:self];
	return self;
}

+ (MiscString *)controllerName
// Returns the controller name.
{
	return [Misc_classControllerName getObjectForClass:self];
}

+ addDocType:(int)tag name:(const char *)name extension:(const char *)ext 
			openSelector:(SEL)openSel saveSelector:(SEL)saveSel
// Adds a new document type that controllers of this class support.  
// The name is the full name of the doc type.  Extension is the file 
// extension for docs of that type.  Tag is for your own use.  open 
// and save selectors take a single MiscDocType as their argument.  
// Either can be NULL if this is a read-only or write-only file type.
{
	List *typeList = [self docTypes];
	
	if (!name || !*name)  {
		NXLogError("[%s addDocType]:  Attempted to add a document type "
					"with no name.  Document types must have names.", 
					[self name]);
		return nil;
	}
	if (!typeList)  {
		typeList = [[List allocFromZone:[self zone]] init];
		[Misc_classDocTypes setObject:typeList forClass:self];
	}

	[typeList addObject:[[MiscDocType allocFromZone:[self zone]] 
				initForControllerClass:self tag:tag name:name extension:ext 
				openSelector:openSel saveSelector:saveSel]];
	return self;
}

+ (List *)docTypes
// Returns the list of document types supported by this class.
{
	return (List *)[Misc_classDocTypes getObjectForClass:self];
}

+ (MiscDocType *)docTypeForTag:(int)tag
// Returns the doc type whose tag is tag.
{
	List *docTypes = [self docTypes];
	int i, c = [docTypes count];
	MiscDocType *type = nil;
	
	for (i=0; i<c; i++)  {
		type = [docTypes objectAt:i];
		if ([type typeTag] == tag)  {
			return type;
		}
	}
	return nil;
}

+ (MiscDocType *)docTypeForName:(const char *)name
// Returns the doc type whose name is name.
{
	List *docTypes = [self docTypes];
	int i, c = [docTypes count];
	MiscDocType *type = nil;
	
	for (i=0; i<c; i++)  {
		type = [docTypes objectAt:i];
		if (([type typeName]) && (name) && 
					(strcmp([type typeName], name) == 0))  {
			return type;
		}
	}
	return nil;
}

+ (MiscDocType *)docTypeForExtension:(const char *)extension
// Returns the doc type whose extension is extension.
// If we find a match, return it.  If we don't find a match, and we have
// an extension-less doc type, return that.
{
	List *docTypes = [self docTypes];
	int i, c = [docTypes count];
	MiscDocType *type = nil;
	const char *curExt = NULL;
	
	for (i=0; i<c; i++)  {
		type = [docTypes objectAt:i];
		curExt = [type typeExtension];
		if ((!extension || !*extension) && (!curExt || !*curExt))  {
			return type;
		}
		if ((curExt) && (extension) && 
					(strcmp(curExt, extension) == 0))  {
			return type;
		}
	}  
	if (extension && *extension)  {
		return [self docTypeForExtension:NULL];
	}
	return nil;
}

+ (List *)getOpenTypesList:(List *)list
// Fills list with the types which support reading from our docTypes list.
// list is returned.  If list is nil, a List object is allocated for it.
{
	List *docTypes = [self docTypes];
	int i, c = [docTypes count];
	MiscDocType *type = nil;
	
	if (!list)  {
		list = [[List allocFromZone:[self zone]] init];
	}
	
	for (i=0; i<c; i++)  {
		type = [docTypes objectAt:i];
		if ([type canOpenType])  {
			[list addObject:type];
		}
	}
	
	return list;
}

+ (List *)getSaveTypesList:(List *)list
// Fills list with the types which support saving from our docTypes list.
// list is returned.  If list is nil, a List object is allocated for it.
{
	List *docTypes = [self docTypes];
	int i, c = [docTypes count];
	MiscDocType *type = nil;
	
	if (!list)  {
		list = [[List allocFromZone:[self zone]] init];
	}
	
	for (i=0; i<c; i++)  {
		type = [docTypes objectAt:i];
		if ([type canSaveType])  {
			[list addObject:type];
		}
	}
	
	return list;
}

+ startUnloading
// We have to deallocate the class variables.
{
	[Misc_classControllerName free];
	[Misc_classDocTypes free];
	
	return self;
}

- initFromFile:(const char *)path manager:aManager
// This is the initializer that will be used normally.  It just calls the DI.
{
	return [self initWithFrameName:NULL fromFile:path manager:aManager];
}

- initWithFrameName:(const char *)theFrameName
// Just call the DI.
{
	return [self initWithFrameName:theFrameName fromFile:NULL manager:nil];
}

- initWithFrameName:(const char *)theFrameName 
			fromFile:(const char *)path manager:aManager
// Designated Initializer.  A null file means a new untitled doc.
{
	[super initWithFrameName:theFrameName];
	
	[self setManager:aManager];
	docName = [[MiscString allocFromZone:[self zone]] init];
	untitledString = [[MiscString allocFromZone:[self zone]] init];
	docType = nil;
	
	isDirty = NO;
	saveAccessoryPanel = nil;
	saveAccessoryBox = nil;
	saveAccessoryPopupButton = nil;
	
	[self setFile:path];
	
	return self;
}

- free
// free our ivars
{
	[docName free];
	[untitledString free];
	if (manager)  [manager removeDocument:self];
	
	return [super free];
}

- nibDidLoad
// Set the window title, move the window according to the frameCycle.
{
	[super nibDidLoad];
	
	if (window)  {
		[self resetWindowTitle];
		if (manager)  {
			NXPoint pt;
			[manager getNextWindowLocation:&pt];
			[window moveTopLeftTo:pt.x :pt.y];
		}
	}
	return self;
}

- (BOOL)doClear
// override this to totally clear this doc into a blank one.
{
	return YES;
}

- (BOOL)doPrint
// override this to print the document (including running print panel, etc...).
{
	return YES;
}

- saveAccessoryPopupAction:sender
// The action of the save accessory document type popup.  It sets 
// the Save panels required file extension appropriately.
{
	id savePanel = [SavePanel new];
	List *saveTypes = [[self class] getSaveTypesList:nil];
	
	[savePanel setRequiredFileType:[[saveTypes objectAt:
				[sender selectedRow]] typeExtension]];
	
	[saveTypes free];
	return self;
}

- getSaveAccessoryForSaveTypes:(List *)saveTypes 
			currentType:(MiscDocType *)currentType
// This will prepare and return the saveAccessory view for the given sender 
// document.  If necessary it will load the accessory nib first.  It returns 
// nil if there is only a single save type since no accessory is needed then.
// Subclasses may override this to use a different nib file.  If a subclass
// does override, it is important that the new nib file NOT use the existing
// saveAccessory... outlets for MiscDocument.  Declare all new outlets for
// your save accessory nib in your subclass and use those.  The saveAccessory
// outlets in this class are not used outside this method, but other subclasses
// may not override, and they should get the old nib.
{
	id popup = nil;
	int i, c;
	MiscString *scratch = nil;
	MiscDocType *type = nil;
	
	c = [saveTypes count];
	if (c <= 1)  {
		return nil;
	}
	
	if (!saveAccessoryBox)  {
		if (MiscKitBundle)  {
			char pathBuff[MAXPATHLEN+1];
			if (![MiscKitBundle getPath:pathBuff 
						forResource:SAVEACCESSORY_NIBNAME ofType:NULL])  {
				NXLogError("%s: failed to find nib file %s in MiscKit bundle.", 
						[[self class] name], SAVEACCESSORY_NIBNAME);
				return nil;
			}
			if (![NXApp loadNibFile:pathBuff owner:self])  {
				NXLogError("%s: failed to load nib file %s.", 
						[[self class] name], pathBuff);
				return nil;
			}
		}  else  {
			if (![NXApp loadNibSection:SAVEACCESSORY_NIBNAME owner:self])  {
				NXLogError("%s: failed to load nib file %s.", 
						[[self class] name], SAVEACCESSORY_NIBNAME);
				return nil;
			}
		}
		[saveAccessoryBox removeFromSuperview];
		[saveAccessoryPanel free];
		saveAccessoryPanel = nil;
	}
	
	popup = [saveAccessoryPopupButton target];
	
	// remove all but the first item in the list
	for (i=[popup count]-1; i>0; i--)  {
		[[popup removeItemAt:i] free];
	}
	
	// set the first item's title to ""
	[[[popup itemList] cellAt:0:0] setTitle:""];
	
	scratch = [[MiscString allocFromZone:[self zone]] init];

	// add the new items
	for (i=0; i<c; i++)  {
		type = [saveTypes objectAt:i];
		[scratch setFromFormat:"%s (%s)", 
					[type typeName], 
					([type typeExtension]?[type typeExtension]:"<none>")];
		[popup addItem:[scratch stringValue]];
		// If this is the current type...
		if (type == currentType)  {
			[saveAccessoryPopupButton setTitle:[scratch stringValue]];
		}
	}
	
	// If there is no current type, select the first one.
	if (currentType == nil)  {
		type = [saveTypes objectAt:0];
		[scratch setFromFormat:"%s (%s)", 
					[type typeName], 
					([type typeExtension]?[type typeExtension]:"<none>")];
		[saveAccessoryPopupButton setTitle:[scratch stringValue]];
	}
	
	// remove the first item (which was still left from the old.
	[[popup removeItemAt:0] free];
	
	// set the target/action
	[popup setTarget:self];
	[popup setAction:@selector(saveAccessoryPopupAction:)];
	
	[popup sizeToFit];
			
	return saveAccessoryBox;
}

- save:sender
// Called from various places when we are supposed to save.
{
	BOOL ret = YES;

	if ([docName emptyString])  {
		return [self saveAs:self];
	}
	ret = ([self perform:[docType saveSelector] with:docType] != nil);
	if (ret)  {
		[self setDirty:NO];
	}
	return ret?self:nil;
}

- saveAs:sender
// Called by DocManager in response to Save As command.
{
	id savePanel = [SavePanel new];
	MiscString *dir, *file, *extension;
	
	List *saveTypes = [[self class] getSaveTypesList:nil];
	id saveAccessory = nil;
	
	if ([saveTypes count] == 0)  {
		[saveTypes free];
		return nil;
	}
	
	saveAccessory = [self getSaveAccessoryForSaveTypes:saveTypes 
			currentType:docType];
	
	[savePanel setAccessoryView:saveAccessory];
	[savePanel setDelegate:self];
	[savePanel setTreatsFilePackagesAsDirectories:NO];
	if (!docType)  {
		[savePanel setRequiredFileType:[[saveTypes objectAt:0] typeExtension]];
	}  else  {
		[savePanel setRequiredFileType:[docType typeExtension]];
	}
	if ([docName emptyString])  {
		dir = [[MiscString allocFromZone:[self zone]] 
					initString:NXHomeDirectory()];
		file = [untitledString copy];
	}  else  {
		file = [docName fileName];
		dir = [docName pathName];
	}
	if (![savePanel runModalForDirectory:[dir stringValue] 
				file:[file stringValue]])  {
		[file free];
		[dir free];
		[saveTypes free];
		return nil;
	}
	
	[docName setStringValue:[savePanel filename]];
	extension = [docName fileExtension];
	docType = [[self class] docTypeForExtension:[extension stringValue]];
	[extension free];
	[self resetWindowTitle];
	[self save:self];
	
	[file free];
	[dir free];
	[saveTypes free];
	return self;
}

- saveTo:sender
// Called by DocManager in response to Save To command.  The bit about 
// the flushWindow is so the title bar doesn't flash the save to path and 
// then reset to the old path visibly.
{
	id copyString = [docName copy];
	BOOL copyDirty = [self isDirty];
	
	if (window)  {
		[window disableFlushWindow];
	}
	[self saveAs:self];
	
	[self setDirty:copyDirty];
	[docName free];
	docName = copyString;
	[self resetWindowTitle];
	
	if (window)  {
		[window reenableFlushWindow];
		[window flushWindow];
	}
	return self;
}

- revert:sender
// Called when a file is opened or reverted.
{
	BOOL ret = YES;
	
	if ([docName emptyString])  {
		ret = [self doClear];
	}  else  {
		ret = ([self perform:[docType openSelector] with:docType] != nil);
	}
	if (ret)  {
		[self resetWindowTitle];
		[self setDirty:NO];
	}
	return ret?self:nil;
}

- close:sender
// Closes the document's window.
{
	[window performClose:self];
	return self;
}

- print:sender
// Prints the doc.
{
	BOOL ret = [self doPrint];
	return ret?self:nil;
}

- saveIfNeeded
// Save ourself only if we are dirty.
{
	if ([self isDirty])  [self save:self];
	return self;
}

- setManager:aManager
// Set the docManager that manages us.
{
	if (manager)  [manager removeDocument:self];
	manager = aManager;
	if (manager)  [manager addDocument:self];
	return self;
}

- manager
// Return our manager.
{
	return manager;
}

- setFile:(const char *)fName
// Name our file.  If fName is NULL, set up an untitled name.  The file
// should be of a type we can read.  Also, if the filename is not of a 
// type we can save, make it untitled after loading.
{
	id extension = nil;
	
	[docName setStringValue:fName];
	if ([docName emptyString])  {
		docType = nil;
		if (manager)  {
			[untitledString setFromFormat:"%s-%d", UNTITLED_STRING, 
						[manager nextUntitledNum]];
		}  else  {
			[untitledString setFromFormat:"%s", UNTITLED_STRING];
		}
	}
	
	[self window:YES];
	if (window)  {
		[window disableFlushWindow];
	}
	if (![docName emptyString])  {
		extension = [docName fileExtension];
		docType = [[self class] docTypeForExtension:[extension stringValue]];
	}
	if (![self revert:self])  {
		NXRunAlertPanel(ERRORALERTPANEL_TITLE, ERRORALERTPANEL_TEXT,
					ERRORALERTPANEL_OK, NULL, NULL, [docName stringValue]);
	}
	
	// If we can't save files of whatever type this is, set us to untitled.
	// This must be done after the revert.
	if (![docName emptyString])  {
		if (!docType || (![docType canSaveType]))  {
			[docName freeString];
			docType = nil;
			if (manager)  {
				[untitledString setFromFormat:"%s-%d", UNTITLED_STRING, 
							[manager nextUntitledNum]];
			}  else  {
				[untitledString setFromFormat:"%s", UNTITLED_STRING];
			}
			[self resetWindowTitle];
		}
	}
	if (extension)  [extension free];
	if (window)  {
		[window reenableFlushWindow];
		[window flushWindow];
	}
	[self showWindow:self];

	return self;
}

- (BOOL)isUntitled
// Returns YES if this doc has no file.
{
	return [docName emptyString];
}

- (const char *)file
// Return our filename (or the untitled name if we have no file).
{
	if (![docName emptyString])  {
		return [docName stringValue];
	}  else  {
		return [untitledString stringValue];
	}
}

- docName
// Return the actual MiscString used to store our doc name.  This method
// can be useful if you want to use the MiscString features.  Otherwise, 
// use the -file method.
{
	return docName;
}

- resetWindowTitle
// Make sure that our window title is correct.
{
	if (window)  [window setTitleAsFilename:[self file]];
	return self;
}

- setDirty:(BOOL)flag
// Mark ourselves as edited.
{
	isDirty = flag;
	if (window)  [window setDocEdited:flag];
	return self;
}

- (BOOL)isDirty
// Returns whether or not the doc is dirty.
{
	return isDirty;
}

- windowDidBecomeMain:sender
// Inform the manager that we should be the current document.
{
	if (manager)  [manager setCurrentDocument:self];
	return self;
}

- windowDidResignMain:sender
// Inform the manager that we should no longer be the current document.
{
	if (manager)  [manager setCurrentDocument:nil];
	return self;
}

- windowWillClose:sender
// If we are dirty, ask if we should save before closing and give the chance
// to cancel.
{
	int ret;
	
	if ([self isDirty])  {
		ret = NXRunAlertPanel(CLOSEALERTPANEL_TITLE, CLOSEALERTPANEL_TEXT,
				      CLOSEALERTPANEL_SAVE, CLOSEALERTPANEL_CLOSEANYWAY,
				      CLOSEALERTPANEL_CANCEL, [self file]);
		if (ret==NX_ALERTDEFAULT)  {
			// save
			if (![self save:self])  {
				return nil;
			}
		}  else if (ret==NX_ALERTALTERNATE)  {
			// close anyway
			// do nothing, just let it close
		}  else  {
			// cancel
			return nil;
		}
	}
	// if we get to here, we are closing
	[NXApp delayedFree:self];
	return self;
}

- awake
// Initialize stuff we don't archive.
{
	saveAccessoryPanel = nil;
	saveAccessoryBox = nil;
	saveAccessoryPopupButton = nil;
	return self;
}

- read:(NXTypedStream *)strm
// This method is probably not useful.  But here it is anyway.
{
	int classVersion;

	[super read:strm];
	
	classVersion = NXTypedStreamClassVersion(strm, CLASS_NAME);
	
	switch (classVersion)  {
		case 0:		// First version.
			docName = NXReadObject(strm);
			untitledString = NXReadObject(strm);
			manager = NXReadObject(strm); // ***** I'm not 100% convinced we want to do this; perhaps we want to write an object reference...
			NXReadType(strm, "C", &isDirty);
			docType = nil;
			[self setFile:[docName stringValue]];
			break;
		default:
			NXLogError("[%s read:] class version %d cannot read "
						"instances archived with version %d", 
						CLASS_NAME, CLASS_VERSION, classVersion);
			docName = [[MiscString allocFromZone:[self zone]] init];
			untitledString = [[MiscString allocFromZone:
						[self zone]] init];
			manager = nil;
			isDirty = NO;
			docType = nil;
			[self setFile:NULL];
			break;
	}
	return self;
}

- write:(NXTypedStream *)strm
// This method is probably not useful.  But here it is anyway.
{
	[super write:strm];
	NXWriteObject(strm, docName);
	NXWriteObject(strm, untitledString);
	NXWriteObjectReference(strm, manager);
	NXWriteType(strm, "C", &isDirty);
	return self;
}

@end

