/*============================ MiscFramingView.m ===========================*/
//		Written by Dale Amon, copyright 1994 by Genesis Project, Ltd.
//				Version 0.8.  All rights reserved.
//
//		This notice may not be removed from this source code.
//
//	This object is included in the MiscKit by permission from the author
//	and its use is governed by the MiscKit license, found in the file
//	"LICENSE.rtf" in the MiscKit distribution.  Please refer to that file
//	for a list of all applicable permissions and restrictions.
//	

/* Class that impliments a frame and background fill. It acts much like a
   subwindow in that there is a frame and a single content view

HISTORY
7-Feb-93   Dale Amon at GPL
	   Created.

NOTE:	We have defined the MiscFramingView such that contents are forced to
	follow its size. So if the content is changed, it will eventually
	get resized back to what it is supposed to be.

TODO:	- finish archive methods
	- add more outline types?
	- disable any other View methods?
	- override any other View methods?
*/

#import	<misckit/misckit.h>
#import	<dpsclient/dpsclient.h>

@implementation MiscFramingView

/*===========================================================================*/
/* Define localized strings */
/*===========================================================================*/

#define ILLEGAL_METHOD NXLocalizedStringFromTable("MiscKit", "  %s objects should not be sent '%s' messages\n", NULL, "Programmer error, not allowed to use the named method")

/*===========================================================================*/
/* Internal methods */
/*===========================================================================*/
/* Set the width needed for the frame part of MiscFramingView */

- _setOutlineWidth
  {
	switch (outline)
	 {case MISC_FV_OUTLINE:	outlineWidth = MISC_FV_OUTLINE_WIDTH;
				break;
	  case MISC_FV_PLAIN:	outlineWidth = MISC_FV_PLAIN_WIDTH;
				break;
	  default:		outlineWidth = MISC_FV_PLAIN_WIDTH;
				outline      = MISC_FV_PLAIN;
				break;
	  }
	return self;
  }

/*---------------------------------------------------------------------------*/
/* Lazy setup. Since lots of things can change the frame, there is no sense
   figuring out what it is until we actually have to draw the damn thing.
   So if the _dirty flag is set, we get the outline width and calculate the
   position and size of the content view within the frame view.
*/

- _lazilyUpdateContentFrame
  {	NXRect	fFrame, cFrame;
	NXCoord fw,fh,cw,ch,scalefactor;
	NXCoord	tw,th;
	BOOL	a,b;

	/*-------------------------------------------------------------------*/
	/* Update the frame boundary width */

	if (!_dirty)	return self;
	[self _setOutlineWidth];

	/*-------------------------------------------------------------------*/
 	if (!contentView) return self;
	[self        getFrame: &fFrame];
	[contentView getFrame: &cFrame];

	/*-------------------------------------------------------------------*/
	/* Resize the contents as required */

	fw = fFrame.size.width  - (2 * outlineWidth);
	fh = fFrame.size.height - (2 * outlineWidth);
	cw = cFrame.size.width;
	ch = cFrame.size.height;

	/* Set up values for the boolean logic equations */
	a  = (cw <= 0.0);	/* width is 0 or invalid */
	b  = (ch <= 0.0);	/* height is 0 or invalid */

	th = (fw/cw) * ch;
	tw = (fh/ch) * cw;

	switch (resizing)
	  {/* resize so that it is the same size as the frame view content area
	      in at least one dimension and leaves the other dimension equal
	      or less: ie the best fit that displays all of the image within
	      the window */
	   case MISC_FV_BESTFIT:
		if (a && b)				scalefactor = 1.0;
		else if (th <= fh)			scalefactor = fw/cw;
		else					scalefactor = fh/ch;

		cFrame.size.width  = scalefactor * cw;
		cFrame.size.height = scalefactor * ch;
		break;

	   /* resize so that it is the same size as the frame view content area
	      in at least one dimension and equal or greater in the other:
	      ie, the best fit that fills the window with the image. */
	   case MISC_FV_FILLFIT:
		if (a && b)				scalefactor = 1.0;
		else if (th >= fh)			scalefactor = fw/cw;
		else					scalefactor = fh/ch;

		cFrame.size.width  = scalefactor * cw;
		cFrame.size.height = scalefactor * ch;
		break;

	   /* Warp it to fit the frame */
	   case MISC_FV_EXACTFIT:	
		cFrame.size.width  = fw;
		cFrame.size.height = fh;
		break;

	   /* Leave it as it is */
	   case MISC_FV_NOFIT:
	   default:		break;
	  }

	/* Resize and get the content frame again, in case the view enforces
	   some rule on what resizing it allows */

	[contentView sizeTo:  cFrame.size.width : cFrame.size.height];
	[contentView getFrame: &cFrame];

	/*-------------------------------------------------------------------*/
	/* Justify content view in MiscFramingView.
	   Position the content view so it is not under the frame. An extra 1
	   is included to account for the top boundary line rounding down a
	   pixel. If trouble persists, try NXIntegralRect().
	*/

	switch (justification)
	  {case MISC_FV_LL:		/* Justify to lower left corner */
				cFrame.origin.x = outlineWidth;
				cFrame.origin.y = outlineWidth;
				break;

	   case MISC_FV_UL:		/* Justify to upper left corner */
				cFrame.origin.x = outlineWidth;
				cFrame.origin.y =
				 fFrame.size.height - 1 - outlineWidth -
				 cFrame.size.height;
				break;

	   case MISC_FV_UR:		/* Justify to upper right corner */
				cFrame.origin.x =
				 fFrame.size.width - outlineWidth -
				 cFrame.size.width;
				cFrame.origin.y =
				 fFrame.size.height - 1 - outlineWidth -
				 cFrame.size.height;
				break;

	   case MISC_FV_LR:		/* Justify to lower right corner */
				cFrame.origin.x =
				 fFrame.size.width - outlineWidth -
				 cFrame.size.width;
				cFrame.origin.y = outlineWidth;
				break;

	   case MISC_FV_CENTER:	/* Justify to center of MiscFramingView */
				cFrame.origin.x =
				 (fFrame.size.width - cFrame.size.width)/2;
				cFrame.origin.y =
				 (fFrame.size.height - cFrame.size.height)/2;

	   default:		break;
	  }

	/*-------------------------------------------------------------------*/
	/* update the content position */

	[contentView moveTo: cFrame.origin.x : cFrame.origin.y];
	_dirty = NO;
	return self;
  }

/*===========================================================================*/
/* Class methods */
/*===========================================================================*/
/* Initialize the class */

+ initialize
{
	[MiscFramingView setVersion:MISC_FRAMINGVIEW_VERSION_ID];
	return self;
}

/*===========================================================================*/
/* Init, alloc and free methods */
/*===========================================================================*/
/* DEFAULT INITIALIZER. Create a window of the specified size with a default
   background gray.
*/

- initFrame: (const NXRect *) theRect
  {	[super initFrame: theRect];
	backgroundGray	= MISC_DEFAULT_BACKGROUND;
	justification   = MISC_DEFAULT_JUSTIFICATION;
	resizing	= MISC_DEFAULT_RESIZING;
	outline		= MISC_DEFAULT_OUTLINE;
	_dirty		= YES;
	return self;
  }

/*---------------------------------------------------------------------------*/
/* discard the pointer to the content view and remove it from the view
   hierarchy since it is not to be destroyed when deallocating the frame
   view.
*/

- free
  {	[self setContentView: nil];
	return [super free];
  }


/*===========================================================================*/
/* set the content view of the MiscFramingView to an ImageView and reposition
   as necessary to make everything fit neatly. aView can be nil. Returns
   the previous contents. Does NOT free the contentView.

   The content view is told to notify us if it changes it's frame for any
   reason.

*/

- setContentView: (View*) aView;
  {	NXRect	cFrame, fFrame;
	id oldContents = contentView;

	[oldContents removeFromSuperview];
	contentView = (id) aView;

	/* If the size is zero, it is a brand new view, so for the moment
	   size it to fit as closely as possible to the frame */
	[contentView getFrame: &cFrame];
	if ((cFrame.size.width == 0) && (cFrame.size.height == 0))
	   {[self        getFrame: &fFrame];
	    [contentView sizeTo: fFrame.size.width : fFrame.size.height];
	   }

	[super addSubview: contentView];
	[contentView notifyAncestorWhenFrameChanged: YES];
	_dirty = YES;
	return oldContents;
  }


/*===========================================================================*/
/* Disallow subview change methods */
/*===========================================================================*/
- addSubview:aView :(int)place relativeTo:otherView
  {	[self error: ILLEGAL_METHOD, [[self class] name], sel_getName(_cmd)];
	return self;
  }

- addSubview:aView
  {	[self error: ILLEGAL_METHOD, [[self class] name], sel_getName(_cmd)];
	return self;
  }

- replaceSubview:oldView with:newView
  {	[self error: ILLEGAL_METHOD, [[self class] name], sel_getName(_cmd)];
	return self;
  }

/*===========================================================================*/
/* Override size change methods */
/*===========================================================================*/
/* Modify the behavior of all size height and width modifying methods to
   set the dirty bit. I included rotations for completeness. However, if the
   sub view is an NXImage, it probably doesn't rotate anyway. Better safe
   than sorry...
 
   The rotateBy and sizeBy methods invoke the rotateTo and sizeTo methods and
   setFrame calls sizeTo: so they do not need to be separately overriden.

*/

- rotateTo:(NXCoord)angle
  {_dirty = YES; return ([super rotateTo:angle]);}

- sizeTo:(NXCoord)width :(NXCoord)height
  {_dirty = YES; return ([super sizeTo: width : height]);}

/*---------------------------------------------------------------------------*/
/* If the content view changes it's size, then set the dirty bit so we will
   correct or accept it's changes.
*/

- descendantFrameChanged:sender
  {	_dirty = YES;
	return ([super descendantFrameChanged:sender]);
  }

/*===========================================================================*/
/* Draw the frame and backdrop. */

- drawSelf:(const NXRect *)rects :(int)rectCount
  {
	[self _lazilyUpdateContentFrame];

	/* Redraw the contents as required */
	PSsetgray (backgroundGray);
	switch (rectCount)
	 {case 1:	NXRectFill(&rects[0]);
			break;
	  case 3:	NXRectFill(&rects[1]);
			NXRectFill(&rects[2]);
			break;
	  default:	break;
	  }

	/* Redraw the frame as required */
	PSsetgray (NX_BLACK);
	switch (outline)
	 {case MISC_FV_OUTLINE:	NXFrameRectWithWidth(&bounds, 
					    (NXCoord)outlineWidth);
				break;
	  default:		break;
	  }

	return self;
   }


/*===========================================================================*/
/* setting and retrieval of ivars */
/*===========================================================================*/

- (float) backgroundGray	{return backgroundGray;}
- (int)   outline		{return outline;}
- setOutline: (int) type	{outline = type;
				 _dirty= YES; return self;}

- (int) justification		{return justification;}
- setJustification: (int) type	{justification = type;
				 _dirty= YES; return self;}

- (int) resizing		{return resizing;}
- setResizing:  (int) type	{resizing  = type;
				 _dirty= YES; return self;}
- contentView			{return contentView;}

/*---------------------------------------------------------------------------*/
/* Set the background gray. If the new value is bad, set the gray at the max
   or min as appropriate.
*/

- setBackgroundGray: (float) grayscale
  {	backgroundGray = grayscale;
	if (backgroundGray > 1.0) backgroundGray = 1.0;
	if (backgroundGray < 0.0) backgroundGray = 0.0;
	return self;
  }


/*===========================================================================*/
/* Archive methods */
/*===========================================================================*/

- write: (NXTypedStream *) stream
  {	[super write: stream];
	NXWriteTypes (stream, "fiii", &backgroundGray, &outline,
				      &justification,&resizing);
	NXWriteObject(stream, contentView);
	return self;
  }

- read: (NXTypedStream *) stream
  {	[super read: stream];
	NXReadTypes  (stream, "fiii", &backgroundGray, &outline,
				      &justification,&resizing);
	contentView = NXReadObject(stream);
	return self;
  }

- awake
  {	[self setOutline: outline];
	return self;
  }

@end
