/*========================= MiscRemoteApplication.m =========================*/
//		Written by Dale Amon, copyright 1994 by Genesis Project, Ltd.
//				Version 0.8.  All rights reserved.
//
//		This notice may not be removed from this source code.
//
//	This object is included in the MiscKit by permission from the author
//	and its use is governed by the MiscKit license, found in the file
//	"LICENSE.rtf" in the MiscKit distribution.  Please refer to that file
//	for a list of all applicable permissions and restrictions.
//	

/* Class for running and setting up communications with remote applications.

HISTORY
25-Feb-93  Dale Amon at GPL
	   Created. A seperate class to encapsulate and generalize the desired
	   behavior of remote application connections. Started from code
	   for gis and scriptplayer connections that was in EMA controller
	   class.

12-Jul-93 Dale Amon at GPL
	  Rewrite of all task control and communications to make it smarter
	  about loading and simpler to follow what is being requested of
	  the remote tasks.

*/

#define APPTYPENAME	"Application"
#define APPDEFVAR	"PATH"
#define APPVAR		"APPLICATIONPATH"
#define APPMODE		(X_OK | R_OK)

#import <misckit/misckit.h>

@implementation MiscRemoteApplication

/*===========================================================================*/
/* Define localized strings */
/*===========================================================================*/

#define ILLEGAL_METHOD NXLocalizedString("  %s objects should not be sent '%s' messages\n", NULL, "Programmer error, not allowed to use the named method")

#define ASK_WSM NXLocalizedString("Not found, asking Workspace Manager...", NULL, "Diagnostic, WSM is being asked about the App, was not found by our own file path search.")

#define APP_NOT_FOUND NXLocalizedString("Could not find %s", NULL, "Diagnostic, WSM could not find the App.")

#define APP_STARTED NXLocalizedString("Started %s", NULL, "Diagnostic, WSM has started the App.")

#define CONNECTED NXLocalizedString("Connected to %s", NULL, "Diagnostic, Connection to remote application succeeded")

#define CONNECT_FAIL NXLocalizedString("Failed connect to %s", NULL,"Diagnostic, Connection to remote application failed")

#define TIMEOUT NXLocalizedString("Timeout on connection to %s\n" , NULL, "Diagnostic, Timed out on attempt to connect to remote application.")

#define LOST_LINK NXLocalizedString("Lost communication with %s %x", NULL, "Diagnostic, Lost communication with a remote application.")

/*===========================================================================*/
/* Class Variables */
/*===========================================================================*/

static MiscFileFinder	*filetype = NULL;
static BOOL	_Misc_RA_debug     = NO;

/*===========================================================================*/
/* Class methods */
/*===========================================================================*/
/* Create a named MiscFileFinder for Applications */

+ initialize
{
	if ( self == [MiscRemoteApplication class] ) {
		[MiscRemoteApplication setVersion: MISC_REMOTEAPPLICATION_VERSION_ID];
		filetype = [[MiscFileFinder alloc]
			      initName: APPTYPENAME
			   defaultPath: NXCopyStringBuffer(getenv(APPDEFVAR))
			  pathVariable: APPVAR
				  mode: APPMODE];
		_Misc_RA_debug    = NO;
	}
	return self;
}


/*---------------------------------------------------------------------------*/
+ (BOOL) debug			{ return _Misc_RA_debug; }
+ setDebug:(BOOL) flg	{ _Misc_RA_debug = flg; return self; }

/*---------------------------------------------------------------------------*/
/* Set the default search path. This path will be used if a file is not found
   on the path defined by the PATH environment variable. If never set, it
   will be the same as PATH.
*/

+ setPath: (const char*) path {[filetype setDefaultPath: path]; return self;}


/*===========================================================================*/
/* Initalize and free methods*/
/*===========================================================================*/
/* DESIGNATED INITIALIZER: for the MiscRemoteApplication class */

#define MAX_RETRY	20
#define RETRY_SECONDS	2

- initAppName: (const char*) app portName: (const char*) port
  {	[super init];

	/* Do not allow creation of a MiscRemoteApplication object without
	   an application name and a port name */
	if (!app || !port) {[self free]; return nil;}

	appName		= NXCopyStringBufferFromZone(app,  [self zone]);
	portName	= NXCopyStringBufferFromZone(port, [self zone]);
	appPath		= NULL;
	proxy		= nil;
	connection	= nil;
	protocol	= nil;
	retryCount	= MAX_RETRY;
	retrySeconds	= RETRY_SECONDS;
	_Misc_RA_debug		= NO;

	return self;
   }


/*---------------------------------------------------------------------------*/
/* Block parent's init method */

- init
  {	[self error:ILLEGAL_METHOD, [[self class] name], sel_getName(_cmd)];
	return self;
  }


/*---------------------------------------------------------------------------*/
/* free whatever we created. Freeing the connection also frees the proxy, 
   if no one else is using it.

   NOTE: what happens to the protocol? is it owned by connection?
*/

- free
  {	free((char*)appName);
	free((char*)portName);

	if (appPath) free((char*)appPath);

	[connection free];
	return [super free];
  }


/*===========================================================================*/
/* Setting up remote processes and communications */
/*===========================================================================*/
/* Return YES if we successfully start the app. The following places are
   searched, in order:

	1) The paths in your environment variable APPLICATIONPATH
	2) The default paths in this Class as set by you with +setPath.
	   If you have never set it, it will be the same as your PATH
	   environment variable
	3) The Workspace manager is asked to search it's secret hidden
	   internal path list.

   The first one found is the one started. A typical situation: you have a
   user copy in a standard place, ie one of the App directories; and you are
   debugging a related set of apps in a set of working directories.

   Use setPath to point to the working directories. If you have not
   set APPLICATIONPATH , the setPATH directory will hide the ones in the
   real application path that the Workspace keeps. If at run time you need
   to override the program default, you can then 
	setenv APPLICATIONPATH <path>

   It returns nil if the application was not accepted by the workspace manager
   for execution.

   NeXT should have better communications with Workspace manager. They need
   to give us a means of inquiring on the status of an App.

   Beware of multiple launches. Multiple background windows with the
   cursor hidden can confuse things to the point where your
   cursor does not make it's reappearance.

*/

- run
  {	const char *tmp;

	if (appPath) {free((char*) appPath); appPath = NULL;}

	appPath = [filetype fullPathForFile: appName];

	if (_Misc_RA_debug && !appPath) NXLogError(ASK_WSM);

	if (!appPath)
	   {tmp = [[Application workspace]
			getFullPathForApplication: appName];
	    if (!tmp) {NXLogError(APP_NOT_FOUND,appName); return nil;}
	    else appPath = NXCopyStringBufferFromZone(tmp, [self zone]);
	   }

	if (![[Application workspace] launchApplication: appPath
					       showTile: NO  autolaunch:YES])
	 return nil;

	if (_Misc_RA_debug) NXLogError(APP_STARTED, appName);
	return self;
  }


/*---------------------------------------------------------------------------*/
/* INTERNAL: Connect to app on the local host. */

- connectToApp
  {
	proxy = [NXConnection connectToName: portName];
	if (_Misc_RA_debug)
	   {if (proxy)	NXLogError(CONNECTED,    portName);
	    else	NXLogError(CONNECT_FAIL, portName);
	   }
	if (!proxy) return nil;

	[proxy setProtocolForProxy: protocol];

	connection = [proxy connectionForProxy];
	[connection registerForInvalidationNotification:self];

	/* This will let the server call us back on it's private
	  port to us. I am assuming this is secure between this app
	  and the target. */

	[connection runFromAppKit];

	return proxy;
  }


/*---------------------------------------------------------------------------*/
/* INTERNAL: Connect to app on the local host. Retry up to the specified number
   of times.
*/

- connectToAppWithRetry
  {	int	retrycnt = 0;

	while (![self connectToApp])
	 {if (retrycnt++ < retryCount) sleep(retrySeconds);
 	  else {NXLogError(TIMEOUT, portName); return nil;}
	 }
	
	return proxy;
  }


/*---------------------------------------------------------------------------*/
/* Return YES if we:.
	- we are already connected to the application
	- the local name server can connect us
	- we can start the application locally and then connect to it
*/

- connection
  {	if (proxy || [self connectToApp])	return proxy;
	if (![self run])			return nil;

	[self connectToAppWithRetry];
	return proxy;
  }

/*---------------------------------------------------------------------------*/

- unconnect
  {
	[NXConnection unregisterForInvalidationNotification: proxy];
	[connection free]; proxy = connection = nil;
	return self;
  }

/*===========================================================================*/
/* Ivar methods */
/*===========================================================================*/

- (int) retryInterval		{return retrySeconds;}
- setRetryInterval: (int) n	{retrySeconds = n;	return self;}
- (int) retries			{return retryCount;}
- setRetries:	    (int) n	{retryCount = n;	return self;}
- delegate			{return delegate;}
- setDelegate:	    anObject	{delegate = anObject;	return self;}
- protocol			{return protocol;}

- setProtocol: aProtocol
  {	protocol = aProtocol;
	[proxy setProtocolForProxy: protocol];
	return self;
  }

/*===========================================================================*/
/* Invalidation protocol */
/*===========================================================================*/
/* SenderIsInvalid Protocol method.
   Sends a senderIsInvalid message on to the delegate if there is one.

	NOTE: Think a bit more deeply on what to do in each case
*/

- senderIsInvalid:sender
 {
	if (sender == connection)
	   {	NXLogError(LOST_LINK, portName, sender);
		proxy = connection = nil;
		[delegate senderIsInvalid:self];
	   }

	return self;
 }


/*===========================================================================*/
/* Archive methods:
			** TO BE WRITTEN **
	Would need a state var so we would know whether to run the remote
	application as part of the awake procedure.
*/

- write: (NXTypedStream *) stream
  {	[super write: stream];
	return self;
  }


- read: (NXTypedStream *) stream
  {	[super read: stream];
	return self;
  }

@end
