//
//	MiscSelectionMatrix -- a class to make the Matrix more functional
//		Written by Georg Tuparev (c) 1994, 1995 by Georg Tuparev, 
// 		EMBL & Academia Naturalis, Heidelberg, Germany.
//				Version 0.5.1 beta  All rights reserved.
//
//		This notice may not be removed from this source code.
//
//		This is a free object!  Contact the author for the latest version.
//		Georg Tuparev, EMBL / BioComputing, Meyerhofstr. 1, 
//			D-69117 Heidelberg, Germany
//			Phone: +49 - 6221 - 387528
//			Fax:   +49 - 6221 - 387517
//		e-mail:  Tuparev@EMBL-Heidelberg.de
//
//	This object is included in the MiscKit by permission from the author
//	and its use is governed by the MiscKit license, found in the file
//	"LICENSE.rtf" in the MiscKit distribution.  Please refer to that file
//	for a list of all applicable permissions and restrictions.
//	

#import <misckit/MiscSelectionMatrix.h>
#import <objc/objc-runtime.h>

#define CLASS_VERSION 		0

@implementation MiscSelectionMatrix

static id localList = nil;

/*****************************************************************************
 *** Local methods
 *****************************************************************************/
- (void)freeSelectionList
{
	if (selectionOrder) {
		[selectionOrder empty];
		[selectionOrder free];
		selectionOrder = nil;
	}
}

- (void)prepareSelectionList
{
	if (selectionOrder) {
		[selectionOrder empty];
	}
	else {
		selectionOrder = [[List allocFromZone:[self zone]] initCount:10];
	}
}

- (void)syncSelection
{
	int i;
	BOOL changed;
	
	[super getSelectedCells:localList];
	
	for (i = 0; i < [localList count]; i++) {
		if ([selectionOrder indexOf:[localList objectAt:i]] == NX_NOT_IN_LIST) {
			[selectionOrder addObject:[localList objectAt:i]];
		}
	}
	
	do {
		changed = NO;
		for (i = 0; i < [selectionOrder count]; i++) {
			if ([localList indexOf:[selectionOrder objectAt:i]]==NX_NOT_IN_LIST) {
				[selectionOrder removeObject:[selectionOrder objectAt:i]];
				changed = YES;
			}
		}
	} while (changed);
	[localList empty];
}

- (void)copySelList:fromList
{
	if (selectionOrder)
		[selectionOrder free];
	selectionOrder = [fromList copyFromZone:[self zone]];
}

/*****************************************************************************
 *** Overwritten methods
 *****************************************************************************/
+ initialize
// Set the class version
{
	if (self == objc_lookUpClass([MiscSelectionMatrix name]))  {
		[self setVersion:CLASS_VERSION];
	}
	// Alloc the local list
	localList = [[List alloc] init];
	
	return self;
}

// Designated initializer override from Matrix.  Sets up our selection stuff.
- initFrame:(const NXRect *)frm mode:(int)aMode prototype:cellId 
	numRows:(int)rowsHigh numCols:(int)colsWide
{
	[super initFrame:frm mode:aMode prototype:cellId numRows:numRows 
		numCols:numCols];
	
	// Sets up the selection list
	[self prepareSelectionList];
	
	return self;
}

// Designated initializer override from Matrix.  Sets up our selection stuff.
- initFrame:(const NXRect *)frm mode:(int)aMode cellClass:factoryId 
			numRows:(int)rowsHigh numCols:(int)colsWide
{
	[super initFrame:frm mode:aMode cellClass:factoryId numRows:numRows 
		numCols:numCols];
	
	// Sets up the selection list
	[self prepareSelectionList];
	
	return self;
}

- copyFromZone:(NXZone *)zone
{
	id obj = [super copyFromZone:zone];
	// Copy the selection list
	[obj copySelList:selectionOrder];
	return obj;
}

- free
{
	[self freeSelectionList];            // Frees the selection
	return [super free];
}

- mouseDown:(NXEvent *)theEvent
{
	// $$$ implement the Control Dragging stuff
	
	[super mouseDown:theEvent];
	
	[self syncSelection];
	return self;
}

- read:(NXTypedStream *)stream
{
	int classVersion;

	[super read:stream];
	
	classVersion = NXTypedStreamClassVersion(stream,[MiscSelectionMatrix name]);
	
	switch (classVersion)  {
		case 0:		// First version.
			selectionOrder = NXReadObject(stream);
			break;
		default:
			NXLogError("[%s read:] class version %d cannot read "
						"instances archived with version %d", 
						[MiscSelectionMatrix name], CLASS_VERSION, classVersion);
			[self prepareSelectionList];
			break;
	}
	return self;
}

- write:(NXTypedStream *)stream
{
	[super write:stream];
	NXWriteObject(stream,selectionOrder);
	return self;
}

/*****************************************************************************
 *** New interface methods
 *****************************************************************************/
- setReorderActionEnabled:(BOOL)aFlag
{
	// $$$
	return self;
}

- (BOOL)reorderActionEnabled
{
	// $$$
	return NO;
}

- addSelectionForCell:aCell
{
	int cellNumb;
	
	if ([cellList indexOf:aCell] == NX_NOT_IN_LIST)
		return nil;
		
	if (aCell && ([selectionOrder indexOf:aCell] == NX_NOT_IN_LIST)) {
		if ( (cellNumb = [cellList indexOf:aCell]) != NX_NOT_IN_LIST) {
			[self setSelectionFrom:cellNumb to:cellNumb anchor:cellNumb lit:YES];
			[selectionOrder addObject:aCell];
		}
	}
	return aCell;
}

- addSelectionAt:(int)row :(int)col
{
	id theCell = [self cellAt:row:col];
	
	return [self addSelectionForCell:theCell];
}

- addSelectionForRow:(int)row
{
	int i;
	int rows = 0;
	int cols = 0;
	id theCell;
	
	[self getNumRows:&rows numCols:&cols];
	if ((row >= 0) && (row < rows)) {
		for (i = row*cols; i < row*cols+cols; i++) {
			theCell = [cellList objectAt:i];
			[self addSelectionForCell:theCell];
		}
		return self;
	}
	else
		return nil;
}

- addSelectionForColumn:(int)col
{
	int i;
	int rows = 0;
	int cols = 0;
	id theCell;

	[self getNumRows:&rows numCols:&cols];
	if ((col >= 0) && (col < cols)) {
		for (i = col; i < rows*cols; i+=cols) {
			theCell = [cellList objectAt:i];
			[self addSelectionForCell:theCell];
		}
		return self;
	}
	else 
		return nil;
}

- addSelectionForCellWithTag:(int)aTag
{
	id theCell = [self findCellWithTag:aTag];
	
	return [self addSelectionForCell:theCell];
}

- deselectCell:aCell
{
	int cellNumb;
	
	if ([cellList indexOf:aCell] == NX_NOT_IN_LIST)
		return nil;
		
	if (aCell && ([selectionOrder indexOf:aCell] != NX_NOT_IN_LIST)) {
		cellNumb = [cellList indexOf:aCell];
		[self setSelectionFrom:cellNumb to:cellNumb anchor:cellNumb lit:NO];
		[selectionOrder removeObject:aCell];
	}
	return aCell;
}

- deselectCellAt:(int)row :(int)col
{
	id theCell = [self cellAt:row:col];
	
	return [self deselectCell:theCell];
}

- deselectRow:(int)row
{
	int i;
	int rows = 0;
	int cols = 0;

	[self getNumRows:&rows numCols:&cols];
	if ((row >= 0) && (row < rows)) {
		for (i = 0; i < cols; i++) {
			[self deselectCellAt:row:i];
		}
		return self;
	}
	else
		return nil;
}

- deselectColumn:(int)col
{
	int i;
	int rows = 0;
	int cols = 0;

	[self getNumRows:&rows numCols:&cols];
	if ((col >= 0) && (col < cols)) {
		for (i = 0; i < rows; i++) {
			[self deselectCellAt:col:i];
		}
		return self;
	}
	else
		return nil;
}

- deselectCellWithTag:(int)aTag
{
	id theCell = [self findCellWithTag:aTag];
	
	return [self deselectCell:theCell];
}

- getOrderedSelectedCells:(List *)aList
{
	int i;

	[self syncSelection];
	if (!aList) 
		aList = [[List alloc] init];
	for (i = 0; i < [selectionOrder count]; i++)
		[aList addObject:[selectionOrder objectAt:i]];
	return aList;
}


@end

