/*========================= MiscTargetActionView.m =========================*/
//		Written by Dale Amon, copyright 1994 by Genesis Project, Ltd.
//				Version 0.8.  All rights reserved.
//
//		This notice may not be removed from this source code.
//
//	This object is included in the MiscKit by permission from the author
//	and its use is governed by the MiscKit license, found in the file
//	"LICENSE.rtf" in the MiscKit distribution.  Please refer to that file
//	for a list of all applicable permissions and restrictions.
//	

/* Abstract superclass that impliments a target action paradeigm for Views.
   This is not quite the same as the Control Class because there is no 
   "value". It allows the view to accept keyboard and mouse events and convert
   them into target/action messages (potentially using a responder chain).
   Note that events in the actionMask will only be recieved if the
   corresponding events are enabled in the window's event mask.

HISTORY
1-Aug-93   Dale Amon at GPL
	   Created.

NOTE:	- what should awake method do?
	- should setFirstResponder remember the state?
		- handle to resign as well?
		- automatically reset view attaches to a window?

*/

#import	<misckit/misckit.h>

#define ACCEPTED_EVENTS (NX_LMOUSEDOWNMASK | NX_LMOUSEUPMASK | \
			 NX_RMOUSEDOWNMASK | NX_RMOUSEUPMASK | \
			 NX_MOUSEMOVEDMASK | NX_LMOUSEDRAGGEDMASK | \
			 NX_RMOUSEDRAGGEDMASK | NX_MOUSEENTEREDMASK | \
			 NX_MOUSEEXITEDMASK | NX_KEYDOWNMASK | NX_KEYUPMASK | \
			 NX_FLAGSCHANGEDMASK)

@implementation MiscTargetActionView

/*===========================================================================*/
/* Define localized strings */
/*===========================================================================*/
#define FIRST_RESPOND NXLocalizedStringFromTable("MiscKit", "Accept first responder", NULL, "Debug message. View recieved acceptFirstResponder message.")

#define FIRST_MOUSE NXLocalizedStringFromTable("MiscKit", "Accept first mouse", NULL, "Debug message. View received acceptFirtsMouse message.")

#define FLAGS_CHANGED NXLocalizedStringFromTable("MiscKit", "Flags Changed", NULL, "Debug message. View recieved flagsChanged message.")

#define KEY_DOWN NXLocalizedStringFromTable("MiscKit", "Key down", NULL, "Debug message. View recieved keyDown message.")

#define KEY_UP NXLocalizedStringFromTable("MiscKit", "Key up", NULL, "Debug message. View recieved keyUp message.")

#define LEFT_MOUSE_DOWN NXLocalizedStringFromTable("MiscKit", "Left Mouse down", NULL, "Debug message. View recieved mouseDown message.")

#define LEFT_MOUSE_DRAG NXLocalizedStringFromTable("MiscKit", "Left Mouse Dragged", NULL, "Debug message. View recieved mouseDragged message.")

#define MOUSE_ENTERED NXLocalizedStringFromTable("MiscKit", "Mouse Entered", NULL, "Debug message. View recieved mouseEntered message.")

#define MOUSE_EXITED NXLocalizedStringFromTable("MiscKit", "Mouse Exited", NULL, "Debug message. View recieved mouseExited message.")

#define MOUSE_MOVED NXLocalizedStringFromTable("MiscKit", "Mouse Moved", NULL, "Debug message. View recieved mouseMoved message..")

#define LEFT_MOUSE_UP NXLocalizedStringFromTable("MiscKit", "Left Mouse Up", NULL, "Debug message. View recieved mouseUp message.")

#define RIGHT_MOUSE_DOWN NXLocalizedStringFromTable("MiscKit", "Right Mouse down", NULL, "Debug message. View recieved rightMouseDown message.")

#define RIGHT_MOUSE_DRAG NXLocalizedStringFromTable("MiscKit", "Right Mouse Dragged", NULL, "Debug message. View recieved rightMouseDragged message.")

#define RIGHT_MOUSE_UP NXLocalizedStringFromTable("MiscKit", "Right Mouse Up", NULL, "Debug message. View recieved rightMouseUp message.")

/*===========================================================================*/
/* Class methods */
/*===========================================================================*/
/* Initialize the class */

+ initialize
  {	[MiscTargetActionView setVersion:MISC_TAVIEW_VERSION_ID];
	return self;
  }

/*===========================================================================*/
/* Internal Instance methods */
/*===========================================================================*/
/* The print name is the class name. */

- (const char *) _pname
  {	const char *who = [[self class] name];
	return who;
  }

/*---------------------------------------------------------------------------*/
/* Print a message to the log in the format:
	imagename: message
*/

- (void) _logMsg: (const char *) what 
  {NXLogError("%s: %s",[self _pname],what);}


/*===========================================================================*/
/* Init, alloc and free methods */
/*===========================================================================*/
/* DEFAULT INITIALIZER. Create a window of the specified size. Action mask
   is empty and there is no target action declared.
*/

- initFrame: (const NXRect *) theRect
  {	[super initFrame: theRect];
	actionMask	= NX_NULLEVENTMASK;
	target		= nil;
	action		= (SEL)0;
	actionEvent	= NULL;
	debug		= NO;

	return self;
  }


/*===========================================================================*/
/* Action mask and first responder handling */
/*===========================================================================*/
/* if any action mask bits for keyboard is set, accept first responder */

- (BOOL)acceptsFirstResponder
  {	int temp = actionMask & 
		(NX_KEYDOWNMASK | NX_KEYUPMASK | NX_FLAGSCHANGEDMASK);
	if (debug) [self _logMsg: FIRST_RESPOND];
	return (temp ? YES : NO);
  }

- (BOOL)acceptsFirstMouse 
  {	if (debug && firstMouse) [self _logMsg: FIRST_MOUSE];
	return firstMouse;
  }

- setFirstMouse: (BOOL) theMouse {firstMouse = theMouse; return self;}

/*---------------------------------------------------------------------------*/
/* This must be done AFTER this view becomes part of a view hierarchy.
   Tells the window to make us the first responder
*/

- setFirstResponder {return ([[self window] makeFirstResponder: self]);}

/*---------------------------------------------------------------------------*/
/* set and clear action mask bits. Protects against setting bits that we don't
   respond to. Logic equivalent to MOV, BIC & BIS.
*/
	
- (int)addToActionMask:(int)newActions
  {	int	oldActions = actionMask;
	actionMask |= (newActions & ACCEPTED_EVENTS);
	return oldActions;
  }

- (int)actionMask {return actionMask;}

- (int)removeFromActionMask:(int)removeActions;
  {	int	oldActions = actionMask;
	actionMask &= ~(removeActions & ACCEPTED_EVENTS);
	return oldActions;
  }

- (int)setActionMask:(int)newActions;
  {	int	oldActions = actionMask;
	actionMask = newActions & ACCEPTED_EVENTS;
	return oldActions;
  }


/*===========================================================================*/
/* Set up for the target and action paradeigm. Note that this differs from
   Control classes in that there is no value. This is only a smarter handler
   for the responder chain.
*/

- setAction: (SEL) theAction	{action = theAction; return self;}
- setTarget: theTarget		{target = theTarget; return self;}
- (SEL) action			{return action;}
- target			{return target;}

- (const NXEvent *) actionEvent {return (const NXEvent *) actionEvent;}


/*===========================================================================*/
/* Override all the Responder event methods */
/*===========================================================================*/
/* If the action is enabled in the actionMask, the target action is sent.
   If the action is not enabled, the target does not accept the action, or
   returns nil, then the Responder version is called instead.

   The target can find out what event triggered the action by:
	[sender actionEvent];

   Note that the actionEvent is only available during the execution of
   the action method.

*/

- _chk: (NXEvent *)evn msk: (int) mask msg: (const char*)msg
  {	actionEvent = evn;
	if (debug) [self _logMsg: msg];
	if ((actionMask & mask) && action &&
	    [self tryToPerform: action with: target])
	   {actionEvent = NULL; return self;}
	else 
	   {actionEvent = NULL; return nil;}
  }
 
- flagsChanged:(NXEvent *)evn
  {return
    ([self _chk: evn msk: NX_FLAGSCHANGEDMASK	msg: FLAGS_CHANGED	]) ?
	self : [super flagsChanged: evn];}
	
- keyDown:(NXEvent *)evn
  {return
    ([self _chk: evn msk: NX_KEYDOWNMASK	msg: KEY_DOWN		]) ?
	self : [super keyDown: evn];}

- keyUp:(NXEvent *)evn
  {return
    ([self _chk: evn msk: NX_KEYUPMASK		msg: KEY_UP		]) ?
	self : [super keyUp: evn];}

- mouseDown:(NXEvent *)evn
  {return
    ([self _chk: evn msk: NX_LMOUSEDOWNMASK	msg: LEFT_MOUSE_DOWN	]) ?
	self : [super mouseDown: evn];}

- mouseDragged:(NXEvent *)evn
  {return
    ([self _chk: evn msk: NX_LMOUSEDRAGGEDMASK	msg: LEFT_MOUSE_DRAG	]) ?
	self : [super mouseDragged: evn];}

- mouseEntered:(NXEvent *)evn
  {return
    ([self _chk: evn msk: NX_MOUSEENTEREDMASK	msg: MOUSE_ENTERED	]) ?
	self : [super mouseEntered: evn];}

- mouseExited:(NXEvent *)evn
  {return
    ([self _chk: evn msk: NX_MOUSEEXITEDMASK	msg: MOUSE_EXITED	]) ?
	self : [super mouseExited: evn];}

- mouseMoved:(NXEvent *)evn
  {return
    ([self _chk: evn msk: NX_MOUSEMOVEDMASK	msg: MOUSE_MOVED	]) ?
	self : [super mouseMoved: evn];}

- mouseUp:(NXEvent *)evn
  {return
    ([self _chk: evn msk: NX_LMOUSEUPMASK	msg: LEFT_MOUSE_UP	]) ?
	self : [super mouseUp: evn];}

- rightMouseDown:(NXEvent *)evn
  {return
    ([self _chk: evn msk: NX_RMOUSEDOWNMASK	msg: RIGHT_MOUSE_DOWN	]) ?
	self : [super rightMouseDown: evn];}

- rightMouseDragged:(NXEvent *)evn;
  {return
    ([self _chk: evn msk: NX_RMOUSEDRAGGEDMASK	msg: RIGHT_MOUSE_DRAG	]) ?
	self : [super rightMouseDragged: evn];}

- rightMouseUp:(NXEvent *)evn
  {return
    ([self _chk: evn msk: NX_RMOUSEUPMASK	msg: RIGHT_MOUSE_UP	]) ?
	self : [super rightMouseUp: evn];}


/*===========================================================================*/
/* Ivar methods */
/*===========================================================================*/
/* A debug flag is available for enabling debug printouts or other diagnostic
   code
*/

- setDebug: (BOOL) flg	{debug = flg; return self;}
- (BOOL) debug		{return debug;}


/*===========================================================================*/
/* Action methods */
/*===========================================================================*/

- toggleDebug: sender {[self setDebug: ~debug]; return self;}

/*===========================================================================*/
/* Archive methods */
/*===========================================================================*/

- write: (NXTypedStream *) stream
  {	[super write: stream];
	NXWriteTypes  (stream, "ii:i", &firstMouse, &actionMask, &action,
		       &debug);
	NXWriteObjectReference(stream, target);
	return self;
  }

- read: (NXTypedStream *) stream
  {	[super read: stream];
	NXReadTypes (stream, "ii:i", &firstMouse, &actionMask, &action,
		     &debug);
	target = NXReadObject(stream);
	return self;
  }

/* NOTE: may need to set first responder here? */

- awake
  {
	return self;
  }

@end
